"""
A module containing equations of motions for common quantities in GEF models.
"""
import numpy as np
import math
from typing import Tuple

def friedmann(*rhos) -> float:
    r"""
    Calculate the Hubble rate $H$ from the Friedmann equation.

    Parameters
    ----------
    rhos
        list of energy densities

    Returns
    -------
    H : float
        the Hubble rate
    """
    Hsq = (1/3) * (sum(rhos)) 
    return np.sqrt(Hsq)

def check_accelerated_expansion(rhos, ps):
    r"""
    Compute $6 M_{\rm P}^2 \ddot{a}/a$.

    Parameters
    ----------
    rhos : list
        energy densities
    ps : list
        pressures
    """
    return -(sum(rhos) + 3*sum(ps))


def klein_gordon(dphi : float, dV : float, H : float, friction : float) -> float:
    r"""
    Calculate the Klein&ndash;Gordon equation (including gauge-field friction).

    Parameters
    ----------
    dphi : float
        the inflaton velocity, $\dot{\varphi}$
    dV : float
        the inflaton potential gradient, $V_{,\varphi}$
    H : float
        the Hubble rate, $H$
    friction : float
        friction term, (e.g., $\beta/M_{\rm P} \langle {\bf E}\cdot {\bf B} \rangle$)

    Returns
    -------
    ddphi : float
        the inflaton acceleration $\ddot{\varphi}$
    """
    return (-3*H * dphi - dV + friction)

def dlnkh(kh : float, dphi : float, ddphi : float,
             dI : float, ddI : float, xi : float,
               a : float, H : float) -> float:
    r"""
    Calculate the ${\rm d} \log k_{\rm h} / {\rm d}t$.

    Parameters
    ----------
    kh : float
        the instability scale $k_{\rm h}$
    dphi : float
        the inflaton velocity, $\dot{\varphi}$
    ddphi : float
        the inflaton acceleration, $\ddot{\varphi}$
    dI : float
        the inflaton--gauge-field coupling, $I_{,\varphi}$
    ddI : float
        derivative of the inflaton--gauge-field coupling, $I_{,\varphi \varphi}$
    xi : float
        the instability parameter, $\xi$
    a : float
        the scale factor, $a$
    H : float
        the Hubble rate, $H$

    Returns
    -------
    dlnkh : float
         ${\rm d} \log k_{\rm h} / {\rm d}t$.
    """
    
    r = 2*abs(xi)
        
    fc = a * H * r
    
    dHdt = 0. #approximation (quasi de Sitter)
    xiprime = (-dHdt * xi + ( ddI*dphi**2 + dI*ddphi)/2)/H
    rprime = 2*np.sign(xi)*xiprime
    fcprime = H*fc + dHdt*a*r + a*H*rprime
                
    return fcprime/kh

def gauge_field_ode(F : np.ndarray, a : float, kh : float, sclrCpl : float,
          W : np.ndarray, dlnkhdt : float, L : int=10) -> np.ndarray:
    r"""
    Calculate the derivative of
    
    $$\mathcal{F}_\mathcal{E}^{(n)} =  \frac{a^4}{k_{\mathrm{h}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf E}\rangle \,  ,$$
    $$\mathcal{F}_\mathcal{B}^{(n)} =  \frac{a^4}{k_{\mathrm{h}}^{n+4}}\langle {\bf B} \cdot \operatorname{rot}^n {\bf B}\rangle \, , $$
    $$\mathcal{F}_\mathcal{G}^{(n)} =  -\frac{a^4}{2 k_{\mathrm{h}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf B} + {\bf B} \cdot \operatorname{rot}^n {\bf E}\rangle \, . $$
    

    Parameters
    ----------
    F : NDArray
        array [$\mathcal{F}_\mathcal{E}^{(n)}$, $\mathcal{F}_\mathcal{B}^{(n)}$, $\mathcal{F}_\mathcal{G}^{(n)}$] in shape (3,ntr)
    a : float
        the scale factor, $a$
    kh : float
        the instability scale $k_{\rm h}$
    sclrCpl : float
        the inflaton gauge-field coupling, $2H\xi$
    W : NDarray
        boundary terms, shape (3,2)
    dlnkhdt : float
        logarithmic derivative of the instability scale, ${\rm d} \log k_{\rm h} / {\rm d}t$
    L : int
        polynomial order for closing ode at ntr

    Returns
    -------
    dFdt : NDArray
        the time derivative of [$\mathcal{F}_\mathcal{E}^{(n)}$, $\mathcal{F}_\mathcal{B}^{(n)}$, $\mathcal{F}_\mathcal{G}^{(n)}$], shape (3,ntr)
    """
    
    FE = F[:,0]
    FB = F[:,1]
    FG = F[:,2]

    scale = kh/a

    W[2,1] = -W[2,1]

    ns = np.arange(0,  FE.shape[0])

    lams = (-1)**ns

    bdrF = (
            dlnkhdt / (4*np.pi**2) 
            * (np.tensordot(np.ones_like(lams), W[:,0], axes=0)
                + np.tensordot(lams, W[:,1], axes=0))
            )

    dFdt = np.zeros_like(bdrF)

    dFdt[:-1,0] = (bdrF[:-1,0] - (4+ns[:-1])*dlnkhdt*FE[:-1]
                    - 2*scale*FG[1:] + 2*sclrCpl*FG[:-1])
    dFdt[:-1,1] = (bdrF[:-1,1] - (4+ns[:-1])*dlnkhdt*FB[:-1]
                    + 2*scale*FG[1:])
    dFdt[:-1,2] = (bdrF[:-1,2] - (4+ns[:-1])*dlnkhdt*FG[:-1]
                    + scale*(FE[1:] - FB[1:]) + sclrCpl*FB[:-1])

    #truncation conditions:
    L = FE.shape[0]//5
    ls = np.arange(1, L+1, 1)
    facl = np.array([math.comb(L, j) for j in range(1,L+1)])
    FEtr = np.sum( (-1)**(ls-1) * facl * FE[-2*ls], axis=0 )
    FBtr = np.sum( (-1)**(ls-1) * facl * FB[-2*ls], axis=0 )
    FGtr = np.sum( (-1)**(ls-1) * facl * FG[-2*ls], axis=0 )

    dFdt[-1,0] = (bdrF[-1,0] -  (4+ns[-1])*dlnkhdt*FE[-1]
                   - 2*scale*FGtr + 2*sclrCpl*FG[-1])
    dFdt[-1,1] = (bdrF[-1,1] - (4+ns[-1])*dlnkhdt*FB[-1]
                   + 2*scale*FGtr) 
    dFdt[-1,2] = (bdrF[-1,2] - (4+ns[-1])*dlnkhdt*FG[-1]
                   + scale*(FEtr - FBtr) + sclrCpl*FB[-1])

    return dFdt

def dlnkh_schwinger(kh : float, dphi : float, ddphi : float,
                dI : float, ddI : float,xieff : float, s : float,
                    a : float, H : float) -> float:
    r"""
    Calculate the ${\rm d} \log k_{\rm h} / {\rm d}t$ in presence of conductivities.

    Parameters
    ----------
    kh : float
        the instability scale $k_{\rm h}$
    dphi : float
        the inflaton velocity, $\dot{\varphi}$
    ddphi : float
        the inflaton acceleration, $\ddot{\varphi}$
    dI : float
        the inflaton--gauge-field coupling, $I_{,\varphi}$
    ddI : float
        derivative of the inflaton--gauge-field coupling, $I_{,\varphi \varphi}$
    xieff : float
        the effective instability parameter, $\xi + \sigma_{\rm B}/(2H)$
    s : float or val
        the effective electric conductivity, $\sigma_{\rm E}/(2H)$
    a : float
        the scale factor, $a$
    H : float
        the Hubble rate, $H$

    Returns
    -------
    dlnkh : float
         ${\rm d} \log k_{\rm h} / {\rm d}t$.
    """
    sqrtterm = np.sqrt(xieff**2 + s**2 + s)
    r = (abs(xieff) + sqrtterm)
    
    fc = a * H * r

    #approximations
    dsigmaEdt = 0.
    dsigmaBdt = 0.
    dHdt = 0.#vals.vals["Hprime"]# #approximation  dHdt = alphaH**2  (slow-roll)

    xieffprime = (-dHdt * xieff + (ddI*dphi**2 + dI*ddphi + a*dsigmaBdt)/2)/H
    sEprime = (-dHdt * s + a*dsigmaEdt/2)/H
    rprime = ( (np.sign(xieff)+xieff/sqrtterm)*xieffprime
                 + sEprime*(s+1/2)/sqrtterm )
    fcprime = H*fc + dHdt*a*r + a*H*rprime
                
    return fcprime/kh

def ddelta(delta : float, sigmaE : float) -> float:
    r"""
    Calculate the derivative of the cumulative electric damping, $\Delta = \exp \left(-\int \sigma_{\rm E} {\rm d} t\right)$.

    Parameters
    ----------
    delta : float
        cumulative electric damping, $\Delta$
    sigmaE : float
        electric conductivity, $\sigma_{\rm E}$

    Returns
    -------
    ddelta : float
        the time derivative of $\Delta$
    """
    return -delta*sigmaE

def drhoChi(rhoChi : float, E : float, G : float,
               sigmaE : float, sigmaB : float, H : float) -> float:
    r"""
    Calculate the derivative of the fermion energy density.

    Parameters
    ----------
    rhoChi : float
        the fermion energy density, $\rho_{\chi}$
    E : float
        the electric field expecation value, $\langle {\bf E}^2 \rangle$
    G : float
        the expectation value of $-\langle {\bf E} \cdot {\bf B} \rangle$
    sigmaE : float
        electric conductivity, $\sigma_{\rm E}$
    sigmaB : float
        magnetic conductivity, $\sigma_{\rm B}$
    H : float
        the Hubble rate, $H$

    Returns
    -------
    float
        the time derivative of rhoChi
    """
    return (sigmaE*E - sigmaB*G - 4*H*rhoChi)

def gauge_field_ode_schwinger(F : np.ndarray, a : float, kh : float, sclrCpl : float,
            sigmaE : float, delta : float, 
            W : np.ndarray, dlnkhdt : float, L : int=10) -> np.ndarray:
    r"""
    Calculate the derivative of
    
    $$\mathcal{F}_\mathcal{E}^{(n)} =  \frac{a^4}{k_{\mathrm{h}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf E}\rangle \,  ,$$
    $$\mathcal{F}_\mathcal{B}^{(n)} =  \frac{a^4}{k_{\mathrm{h}}^{n+4}}\langle {\bf B} \cdot \operatorname{rot}^n {\bf B}\rangle \, , $$
    $$\mathcal{F}_\mathcal{G}^{(n)} =  -\frac{a^4}{2 k_{\mathrm{h}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf B} + {\bf B} \cdot \operatorname{rot}^n {\bf E}\rangle \, , $$

    in the presence of Schwinger conductivities.

    Parameters
    ----------
    F : NDArray
        array [$\mathcal{F}_\mathcal{E}^{(n)}$, $\mathcal{F}_\mathcal{B}^{(n)}$, $\mathcal{F}_\mathcal{G}^{(n)}$] in shape (3,ntr)
    a : float
        the scale factor, $a$
    kh : float
        the instability scale $k_{\rm h}$
    sclrCpl : float
        the inflaton gauge-field coupling, $2H\xi_{\rm eff}$
    delta : float
        cumulative electric damping, $\Delta$
    W : NDarray
        boundary terms, shape (3,2)
    dlnkhdt : float
        logarithmic derivative of the instability scale, ${\rm d} \log k_{\rm h} / {\rm d}t$
    L : int
        polynomial order for closing ode at ntr

    Returns
    -------
    dFdt : NDArray
        the time derivative of [$\mathcal{F}_\mathcal{E}^{(n)}$, $\mathcal{F}_\mathcal{B}^{(n)}$, $\mathcal{F}_\mathcal{G}^{(n)}$], shape (3,ntr)
    """
    FE = F[:,0]
    FB = F[:,1]
    FG = F[:,2]

    scale = kh/a

    W[2,1] = -W[2,1]

    ns = np.arange(0,  FE.shape[0])

    lams = (-1)**ns

    bdrF = (dlnkhdt*delta / (4*np.pi**2)
            * (np.tensordot(np.ones_like(lams), W[:,0], axes=0)
                + np.tensordot(lams, W[:,1], axes=0))
            )

    dFdt = np.zeros(bdrF.shape)

    dFdt[:-1,0] = (bdrF[:-1,0] - (4+ns[:-1])*dlnkhdt*FE[:-1]
                   - 2*sigmaE*FE[:-1] - 2*scale*FG[1:] + 2*sclrCpl*FG[:-1])
    dFdt[:-1,1] = (bdrF[:-1,1] - (4+ns[:-1])*dlnkhdt*FB[:-1]
                    + 2*scale*FG[1:])
    dFdt[:-1,2] = (bdrF[:-1,2] - (4+ns[:-1])*dlnkhdt*FG[:-1]
                   - sigmaE*FG[:-1] + scale*(FE[1:] - FB[1:]) + sclrCpl*FB[:-1])
    
    #truncation conditions:
    ls = np.arange(1, L+1, 1)
    facl = np.array([math.comb(L, j) for j in range(1,L+1)])
    FEtr = np.sum( (-1)**(ls-1) * facl * FE[-2*ls], axis=0 ) #-2*ls instead of -2*ls+1 since -1 is ntr not ntr-1
    FBtr = np.sum( (-1)**(ls-1) * facl * FB[-2*ls], axis=0 )
    FGtr = np.sum( (-1)**(ls-1) * facl * FG[-2*ls], axis=0 )

    #bilinears at truncation order ntr
    dFdt[-1,0] = (bdrF[-1,0] - (4+ns[-1])*dlnkhdt*FE[-1]
                   - 2*sigmaE*FE[-1] - 2*scale*FGtr + 2*sclrCpl*FG[-1])
    dFdt[-1,1] = (bdrF[-1,1] - (4+ns[-1])*dlnkhdt*FB[-1]
                   + 2*scale*FGtr) 
    dFdt[-1,2] = (bdrF[-1,2] - (4+ns[-1])*dlnkhdt*FG[-1]
                   - sigmaE*FG[-1] + scale*(FEtr - FBtr) + sclrCpl*FB[-1])

    return dFdt

def conductivities_collinear(a  : float, H  : float,
                           E  : float, B : float, G : float,
                             picture : int, omega : float) -> Tuple[float, float, float]:
    r"""
    Compute electric & magnetic conductivities and the damping scale $k_{\rm S}$
    assuming collinear E & M fields.

    Parameters
    ----------
    a : float
        the scale factor, $a$
    H : float
        the Hubble rate, $H$
    E : float
        the electric field expecation value, $\langle {\bf E}^2 \rangle$
    E : float
        the magnetic field expecation value, $\langle {\bf B}^2 \rangle$
    G : float
        the expectation value of $-\langle {\bf E} \cdot {\bf B} \rangle$
    picture : int
        an integer specifying electric (=-1) or magnetic pictures (=1) 
    omega : float
        the reference frequency to convert from numerical to physical units

    Returns
    -------
    float
        the electric damping, $\sigma_{\rm E}$
    float
        the magnetic damping, $\sigma_{\rm B}$
    float
        the damping scale, $k_{\rm S}$
    """     
    mu = (E+B)
    if mu<=0:
        return 0., 0., 1e-4*a*H
    else:
        mu = (mu/2)**(1/4)
        mz = 91.2/(2.43536e18)
        gmz = 0.35
        gmu = np.sqrt(gmz**2/(1 + gmz**2*41./(48.*np.pi**2)*np.log(mz/(mu*omega))))
        
        C = 41/12

        sigma = (C*gmu**3/(6*np.pi**2 * H * np.tanh(np.pi*np.sqrt(B/E))))
        sigmaE =  np.sqrt(B) * (min(1., (1.- picture))*E + max(-picture, 0.)*B) * sigma / (E+B)         
        sigmaB = -np.sign(G) * np.sqrt(E)*(min(1., (1.+ picture))*B + max(picture,0.)*E)* sigma/(E+B)
        
        ks = C**(1/3)*gmu*E**(1/4)*a
        
        return sigmaE, sigmaB, ks
    
def conductivities_mixed(a  : float, H  : float,
                           E  : float, B : float, G : float,
                            omega : float) -> Tuple[float, float, float]:
    r"""
    Compute electric & magnetic conductivities and the damping scale $k_{\rm S}$
    in the mixed picture.

    Parameters
    ----------
    a : float
        the scale factor, $a$
    H : float
        the Hubble rate, $H$
    E : float
        the electric field expecation value, $\langle {\bf E}^2 \rangle$
    E : float
        the magnetic field expecation value, $\langle {\bf B}^2 \rangle$
    G : float
        the expectation value of $-\langle {\bf E} \cdot {\bf B} \rangle$
    omega : float
        the reference frequency to convert from numerical to physical units

    Returns
    -------
    float
        the electric damping, $\sigma_{\rm E}$
    float
        the magnetic damping, $\sigma_{\rm B}$
    float
        the damping scale, $k_{\rm S}$
    """     
    Sigma = np.sqrt((E - B)**2 + 4*G**2)
    if Sigma<=0:
        return 0., 0., 1e-4*a*H
    else:
        mz = 91.2/(2.43536e18)
        mu = ((Sigma)/2)**(1/4)
        gmz = 0.35
        gmu = np.sqrt(gmz**2/(1 + gmz**2*41./(48.*np.pi**2)*np.log(mz/(mu*omega))))

        Eprime = np.sqrt( (E - B + Sigma)/2 )
        Bprime = np.sqrt( (B- E + Sigma)/2 )
        Sum = E + B + Sigma

        C = 41/12
        
        sigma = ( C*gmu**3/(6*np.pi**2) / (np.sqrt(Sigma*Sum)*H * np.tanh(np.pi*Bprime/Eprime)))
        
        ks = C**(1/3)*gmu*Eprime**(1/2)*a

        return 2**(1/2)*abs(G)*Eprime*sigma, -2**(1/2)*G*Bprime*sigma, ks




CustomGEFSolver_docs = """
        A modified subclass of `BaseGEFSolver` for a custom GEF model.

        This class overwrites the following class members:
        - `vals_to_yini`
        - `update_vals`
        - `timestep`
        - `known_variables`
        - `known_events`
        """

DOCS = {
    "module":"""
    This module defines the main class for an ODE solver used by GEF models.

    The `BaseGEFSolver` is the the primary class supplied by this module. It defines an algorithm by which the equations of motion for a GEF model are solved.
    This makes use of the `geff.bgtypes.BGSystem` module to simplify conversions between numerical and physical units.

    While the `BaseGEFSolver`is solving the ODE's it can track one or multiple `Event` objects. These events correspond to certain conditions, for example, the end of inflation.
    Occurrences of these events can be designed to influence the solver. For example, an 'end of inflation' `Event` may check if
     the solver has reached the end of inflation, and terminate it if it has.

    The `BaseGEFSolver` class can be used to configure customized solvers using the class factory `GEFSolver` to adapt it to a specific GEF model.
    """,
    
    "BaseGEFSolver":r"""
    A class used to solve the equations of motion defined by a GEF model.

    The main purpose of the class is to provide the `compute_GEF_solution` method to `run`. 
    Internally, the method uses `solve_eom`, which wraps `scipy.integrate.solve_ivp`.
    
    All specifications on the GEF model are encoded in the attributes `known_variables` and `known_events`, as well as
    the methods `vals_to_yini`, `update_vals` and `timestep`. 
    For illustration, purposes, these methods are configured such that the GEFSolver solves the
    EoMs for de Sitter expansion:
    $$\frac{{\rm d} \log a}{{\rm d} t} = H_0$$

    In practice, the class factory `GEFSolver` creates a subclass of `BaseGEFSolver` adapted to the EoMs of other models.
    """,

    "GEFSolver":r"""
    Create a subclass of `BaseGEFSolver` with custom equation of motions and initial conditions.

    The subclass is adjusted to the specific EoMs of a new GEF model by defining new methods for `BaseGEFSolver.vals_to_yini`, `BaseGEFSolver.update_vals` and `BaseGEFSolver.timestep` via `new_init`, `new_update_val` and `new_timestep`.
    Information about the classificiation of variables is encoded in `new_variables` which overwrites `BaseGEFSolver.known_variables`.
    
    The `new_init` needs to obey the following rules:
    1. The call signature and output matches `vals_to_yini`.
    2. The return array `yini` has a shape: $n_{\rm dynam.} + 3(n_{\rm tr}+1) \times n_{\rm gauge}$.
    3. All dynamical variables are reflected in the first $n_{\rm dynam.}$-indices of `yini`.
    4. All gauge variables are reflected in the last $3(n_{\rm tr}+1) \times n_{\rm gauge}$-indices of `yini`.

    Above, $n_{\rm dynam.}$ and $n_{\rm gauge}$ are, respectively, the number of dynamical and gauge-field variables.

    The `new_update_vals` needs to obey the following rules:
    1. The call signature and output matches `update_vals`.
    2. It updates every static variable using values in `y` and `t`.

    The `new_timestep` needs to obey the following rules:
    1. The call signature and output matches `timestep`.
    2. It computes derivatives `dydt` for every dynamical or gauge varibale.
    3. Static variables from `new_update_vals` can be re-used, as it is called before `new_timestep`.
    4. The indices in `dydt` need to match those of `yini` returned by `new_init`.

    All these functions assume that `vals` is in numerical units throughout the computation.

    In addition, a new list of `Event` objects can be passed to the subclass using `new_events`

    For an example on how to define a new solver, see `GEFF.models.classic`.

    Parameters
    ----------
    new_init : Callable
        overwrites `BaseGEFSolver.vals_to_yini`
    new_update_vals : Callable
        overwrites `BaseGEFSolver.update_vals`
    new_timestep : Callable
        overwrites `BaseGEFSolver.timestep`
    new_events :  list of Event
        overwrites `BaseGEFSolver.known_events`
    new_variables :  dict
        overwrites `BaseGEFSolver.known_variables`

    Returns
    -------
    GEFSolver
        a subclass of BaseGEFSolver
    """,

    "Event":r"""
    An event which is tracked while solving the GEF equations.

    The class defines a function $f(t, y)$ which is used by `scipy.integrate.solve_ivp` to track occurrences of $f(t, y(t))=0$.
    The event can be `terminal` causing the solver to stop upon an event occurrence.
    The event only triggers if the event condition changes sign according to:
    - positive zero crossing: `direction=1`
    - negative derivative, `direction=-1` 
    - arbitrary zero crossing `direction=0`

    The zeros are recorded and returned as part of the solvers output.

    The function $f$ is encoded in the method `event_func` which is defined upon initialization.

    Within subclasses of `BaseGEFSolver` class, three subclasses of `Event` are used:
    1. `TerminalEvent`
    2. `ErrorEvent`
    3. `ObserverEvent` 
    """,

    "GEFSolver.GEFSolver":CustomGEFSolver_docs
}


import numpy as np

from geff._docs.docs_gw import DOCS
from geff.utility.cosmo import g_rho, g_rho_freq, g_rho_0, g_s, g_s_freq, g_s_0, T_0, M_pl, gev_to_hz, omega_r, h, feq

from numpy.typing import ArrayLike
from typing import Tuple

__doc__ = DOCS["module"]
        
def omega_gw(k:np.ndarray, PT:np.ndarray, Nend:float, Hend:float, Trh:None|float=None) -> Tuple[np.ndarray, np.ndarray]:
    r"""
    Compute $h^2 \Omega_{\rm GW}(f)$ from a tensor power spectrum.

    Parameters
    ----------
    k : NDarray
        momenta in Planck units
    PT : NDarray
        the tensor power spectrum at the end of inflation as a function of momentum
    Nend : float
        the number of e-folds at the end of inflation
    Hend : float
        the Hubble rate at the end of inflation (in Planck units)
    Trh : None or float
        the reheating temperature in GeV. If None, instantaneous reheating is assumed.

    Returns
    -------
    f : NDarray
        frequencies today (in Hz)
    h2OmegaGw : NDarray
        the gravitational-wave spectrum as a function of frequency today
    """

    f = k_to_f(k, Nend, Hend, Trh)
    if Trh is None:
        TransferRH=1
    else:
        frh = 1/(2*np.pi) * (g_s_0/g_s(Trh))**(1/3) * (np.pi**2*g_rho(Trh)/90)**(1/2) * (Trh/M_pl) * T_0*gev_to_hz
        fend = 1/(2*np.pi) * (g_s_0/g_s(Trh))**(1/3) * (np.pi**2*g_rho(Trh)/90)**(1/3) * (Trh/M_pl)**(1/3) * (Hend)**(1/3) * T_0*gev_to_hz

        TransferRH = 1/(1. - 0.22*(f/frh)**1.5 + 0.65*(f/frh)**2)
        TransferRH = np.where(np.log(f/fend) < 0, TransferRH, np.zeros(f.shape))

    TransferMD = 1 + 9/32*(feq/f)**2

    h2OmegaGW = h**2*omega_r/24  * PT * (g_rho_freq(f)/g_rho_0) * (g_s_0/g_s_freq(f))**(4/3) * TransferMD * TransferRH

    return f, h2OmegaGW

def k_to_f(k:np.ndarray, Nend:float, Hend:float, Trh:None|float=None) -> ArrayLike:
    r"""
    Compute frequency today from momentum at the end of inflation

    Parameters
    ----------
    k : NDarray
        momenta in Planck units
    Nend : float
        the number of e-folds at the end of inflation
    Hend : float
        the Hubble rate at the end of inflation
    Trh : None or float
        the reheating temperature in GeV. If None, instantaneous reheating is assumed.

    Return
    ------
    f : NDarray
        frequencies today (in Hz)
    """

    if Trh is None:
        Trh = np.sqrt(3*Hend/np.pi)*(10/106.75)**(1/4)*M_pl
        Trh = Trh*(106.75/g_rho(Trh))**(1/4)
        Nrh = 0
    else:
        wrh = 0
        Nrh = np.log( 90*(Hend*M_pl**2)**2 / (np.pi**2*g_rho(Trh)*Trh**4 ) ) / ( 3 * (1 + wrh) )


    f = k*M_pl*gev_to_hz/(2*np.pi*np.exp(Nend)) * T_0/Trh * (g_s_0/g_s(Trh))**(1/3) * np.exp(-Nrh)

    return f



import numpy as np
from scipy.interpolate import CubicSpline, PchipInterpolator
from scipy.integrate import solve_ivp, trapezoid

from geff._docs.docs_pt import DOCS
from geff.mbm import GaugeSpec
from geff.bgtypes import BGSystem, Constant

from typing import Tuple
from types import NoneType

__doc__ = DOCS["module"]

class PowSpecT:
    r"""
    A class used to compute the tensor power spectrum including vacuum and gauge-field induced contributions.

    These main method of this module is `compute_pt`, which computes both the vacuum and gauge-field induced contribution
      to the tensor power spectrum, $\mathcal{P}_{T,\lambda}^{\mathrm{vac}}$ and $\mathcal{P}_{T,\lambda}^{\mathrm{ind}}$.

    Results are internally computed using numerical units, but are returned in physical units.
    """
    def __init__(self, insys : BGSystem):
        """
        Initialise the class from a GEF solution.

        Parameters
        ----------
        insys : BGSystem
            the GEF solution.
        """
        #Set GEF results to Hubble units.
        sys = BGSystem.from_system(insys, True)
        sys.units = False
        
        #import the background evolution
        N = sys.N.value
        a = np.exp(N)
        H = sys.H.value

        self._omega = sys.omega
            
        #Set the range of modes
        self.maxk = max(a*H)
        self.mink = 1e4

        #Define Useful quantities

        self._t = sys.t.value
        self._N = N
        self._H = H
        self._xi= sys.xi.value

        #define interpolated quantities as needed
        self._af = CubicSpline(self._t, a)
        if isinstance(self._H, Constant):
            self._Hf = lambda t: self._H
        else:
            self._Hf = CubicSpline(self._t, H)
        self._HN = CubicSpline(self._N, self._H)
        self._khN = CubicSpline(N, sys.kh)

        #Obtain eta as a functio of time for phases
        def deta(t, y): return 1/self._af(t)
        
        soleta = solve_ivp(deta, [min(self._t), max(self._t)], np.array([0]), t_eval=self._t)

        self._etaf = CubicSpline(self._t, soleta.y[0,:])
        return
    
    def compute_pt(self, nmodes : int, spec : GaugeSpec, FastGW : bool=True,
                    atols : list=[1e-3,1e-20], rtols : list=[1e-4,1e-4], momgrid : int=100
                    ) -> Tuple[np.ndarray,dict]:
        r"""
        Compute the full tensor power spectrum.

        The method lodes data on gauge modes $A_\lambda(t,k)$ from a file indicated by `mbm_file`.
        The power spectrum is evaluated fpr `nmodes` log-spaced momenta $k \in [10^{4}a(0)H(0), 10 a(t_{\rm max}) H(t_{\rm max})]$

        Parameters
        ----------
        nmodes : NDArray
            the number of momenta
        ModePath : str
            path to a file containing gauge-field mode functions
        FastGW : bool
            If `True`, only the expected dominant contributions to $\mathcal{P}_{T,\lambda}^{\mathrm{ind}}$ are computed
        atols : list
            absolute tolerance for `compute_homogeneous` (index 0) and `compute_green` (index 1)
        rtols : list
            relative tolerance for `compute_homogeneous` (index 0) and `compute_green` (index 1)
        momgrid : int
            passed to `compute_ptind`
        
        Returns
        -------
        ks : NDArray
            the momenta $k$
        PT : dict
            a dictionary containing all contributions to the tensor power spectrum and its total.
        """

        k = np.logspace(np.log10(self.mink), np.log10(10*self.maxk), nmodes)
        ks, tstarts = self._find_tinit_bd(k, mode="k")
        
        Ngrid = spec["N"]
        tgrid = spec["t"]
        pgrid = spec["k"]

        GaugeModes = {"+":(spec["Ap"], spec["dAp"]), "-":(spec["Am"], spec["dAm"])}
        
        indend = np.argmin(abs(Ngrid-max(self._N)))

        PT = {"tot":[], "vac":[], "ind+,++":[], "ind+,+-":[], "ind+,--":[], "ind-,++":[], "ind-,+-":[], "ind-,--":[]}

        GWpols = [("+", 1), ("-",-1)]

        gaugepols=[(("+",1),("+",1)),
                    (("+",1),("-",-1)),
                        (("-",-1),("-",-1))]

        sign = np.sign(self._xi[0])

        for i, k in enumerate(ks):
            tstart = tstarts[i]

            if k > 10**(5/2)*self.maxk:
                for key in PT.keys():
                    PT[key].append(0)
            else:
                uk, _ = self.compute_homogeneous(k, tstart, tgrid, atol=atols[0], rtol=rtols[0])
                Green = self.compute_green(k, uk, tstart, tgrid, indend, atol=atols[1], rtol=rtols[1])
                
                PT["vac"].append(self.compute_ptvac(k, uk[indend]))

                for lgrav in GWpols:
                    for mu in gaugepols:
                        GWpol = lgrav[1]

                        Ax = GaugeModes[mu[0][0]][0]
                        dAx = GaugeModes[mu[0][0]][1]
                        lx = mu[0][1]
                        Ay = GaugeModes[mu[1][0]][0]
                        dAy = GaugeModes[mu[1][0]][1]
                        ly = mu[1][1]

                        if FastGW and (lx != sign or ly != sign):
                            PT[f"ind{lgrav[0]},{mu[0][0]}{mu[1][0]}"].append(0.)
                        else:
                            PT[f"ind{lgrav[0]},{mu[0][0]}{mu[1][0]}"].append(
                                self.compute_ptind(k, GWpol, Ngrid, indend, Green, pgrid, lx, Ax, dAx, ly, Ay, dAy, momgrid) )

        PT["tot"] = np.zeros(ks.shape)
        for key in PT.keys():
            PT[key] = np.array(PT[key])
            if ("+" in key) or ("-" in key):
                PT["tot"] += 0.5*PT[key]
            elif key=="vac":
                PT["tot"] += PT[key]

        #this returns k with units restored. Needs to be matched by omega_GW
        return ks*self._omega, PT
    
    def compute_homogeneous(self, k : float, tvac : float, teval : np.ndarray|NoneType, atol : float=1e-3, rtol : float=1e-4) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Compute the evolution of a vacuum mode starting from Bunch&ndash;Davies vacuum.

        For a mode $u_0(t,k)$ with momentum $k$, initialise in Bunch&ndash;Davies when $k = 10^{5/2} a(t_{\rm vac})H(t_{\rm vac})$.
        Its evolution is obtained by solving `tensor_mode_eq` using `scipy.integrate.solve_ivp`.

        Parameters
        ----------
        k : float
           the momentum $k$.
        tvac : float
            the initial time, $t_{\rm init}$
        teval : NDArray or None
            time coordinates $t$ of $u_0(t,k)$ (`None`: same as class input `sys`) 
        atol : float
            the absolute tolerance for `solve_ivp`
        rtol : float
            the relative tolerance for `solve_ivp`

        Returns
        -------
        u : NDArray
            the vacuum tensor modes, $\sqrt{2k}u(t, k)$
        duk : NDArray
            the derivative of the vacuum tensor modes, $a\sqrt{2/k}\dot{u}(t, k)$
        """

        if len(teval)==0:
            teval = self._t
        tend = max(teval)

        #conformal time needed for relative phases
        eta = self._etaf(teval)

        istart = 0
        while teval[istart]<tvac:
            istart+=1
        
        #define the ODE for the GW modes
        def ode(t, y): return self.tensor_mode_eq( y, k, self._Hf(t), self._af(t) )
        
        #Initialise the modes in Bunch Davies
        Zini = np.array([1, -10**(-5/2), 0, -1])

        sol = solve_ivp(ode, [tvac, tend], Zini, t_eval=teval[istart:], method="RK45", atol=atol, rtol=rtol)
        if not(sol.success):
            print("Something went wrong")

        #the mode was in vacuum before tvac
        vac = list( np.exp(-1j*eta[:istart]*k) )
        dvac = list( -1j*np.exp(-1j*eta[:istart]*k) )

        #Create an array tracking a modes evolution from Bunch Davies to late times. Ensure equal length arrays for every mode k
        uk = np.array( vac + list( (sol.y[0,:] + 1j*sol.y[2,:])*np.exp(-1j*k*eta[istart]) ) )/self._af(teval)
        duk = np.array( dvac + list( (sol.y[1,:] + 1j*sol.y[3,:])*np.exp(-1j*k*eta[istart]) ) )/self._af(teval)

        return uk, duk
    
    @staticmethod
    def tensor_mode_eq(y : np.ndarray, k : float, H : float, a : float) -> np.ndarray:
        r"""
        Mode equation for vacuum tensor modes.

        Parameters
        ----------
        y : NDArray
            the vacuum tensor mode and its derivatives.
            $a \sqrt{2k} u_0(t,k)$ and $ a^2\sqrt{2/k} \dot{u}_0(t, k)$ 
        k : float
            comoving momentum $k$
        H : float
            Hubble rate, $H(t)$
        a : float
            scale factor, $a(t)$

        Returns
        -------
        dydt : NDArray
            an array of time derivatives of y

        """
        dydt = np.zeros(y.shape)

        #real
        dydt[0] = (H*y[0] + k/a*y[1])
        dydt[1] = ( -H*y[1] - k/a*y[0] )

        #imaginary
        dydt[2] = (H*y[2] + k/a*y[3])
        dydt[3] = ( -H*y[3] - k/a*y[2] )

        return dydt
    
    def compute_green(self, k : float, uk : np.ndarray, tvac : float, teval : np.ndarray|NoneType, ind : int, atol : float=1e-30, rtol : float=1e-4) -> np.ndarray:
        r"""
        Compute the evolution of the Green function $G_k(t',t)$ for fixed $t'$.

        The evolution is obtained by solving `green_ode` from $t=t'$ backwards until $t_{\rm vac}$, defined through $k = 10^{5/2} a(t_{\rm end})H(t_{\rm vac})$.
        From $t_{\rm end}$ onwards, the Green function is instead computed from the mode $u_0(t, k)$. The method uses `scipy.integrate.solve_ivp`.

        Parameters
        ----------
        k : float
           the momentum $k$.
        uk : NDArray
            the mode function $u_0(t,k)$
        tvac : float
            the final time, $t_{\rm vac}$
        teval : NDArray or None
            time coordinates $t$ of $G_k(t',t)$ (`None`: same as class input `sys`) 
        ind : int
            index of teval corresponding to $t'$
        atol : float
            the absolute tolerance for `solve_ivp`
        rtol : float
            the relative tolerance for `solve_ivp`

        Return
        ------
        GreenN : NDArray
            the Green function $-k G_k(t', t)$ 
        """
        if len(teval)==0:
            teval = self._t
            
        istart = 0
        while teval[istart]<tvac:
            istart+=1

        # G(k, t, t) = 0 by definition
        Aini = np.array([0, 1])
    

        # Solve the EoM for G backwards in time starting from G(k, t, t)
        def Aode(t, y): return -self.green_ode(y, k, self._Hf(-t), self._af(-t))
        solA = solve_ivp(Aode, [-teval[ind], -tvac], Aini, t_eval=-teval[istart:ind+1][::-1],
                         method="RK45", atol=atol, rtol=rtol)

        #For numerical stability, only solve the EoM for the Green function until t' = tvac. Afterwards, compute it directly from the vacuum modes.
        GreenN = np.zeros(teval.shape)
        GreenN[istart:ind+1] = solA.y[0,:][::-1]
        GreenN[:istart] = ( (uk[ind].conjugate()*uk).imag*self._af(teval)**2 )[:istart]

        return GreenN
    
    @staticmethod
    def green_ode(y : np.ndarray, k : float, H : float, a : float) -> np.ndarray:
        r"""
        Evolve the re-scaled Green function for tensor modes in time.

        The re-scaled Green function is
        $$B_{t'}^k(t) = 2 k a(t)^2 \operatorname{Im} \left[ u_0^*(t', k) \, u_0(t,k) \right] = k G_k(t', t) \,.$$
        
        whose ODE is coupled to 

        $$C_{t'}^k(t) = 2 a(t)^3 \operatorname{Im} \left[ u_0^*(t', k) \, {\dot{u}_0}(t,k) \right] \,.$$

        Parameters
        ----------
        y : NDArray
            the functions B and C
        k : float
            comoving momentum $k$
        H : float
            Hubble rate, $H(t)$
        a : float
            scale factor, $a(t)$

        Returns
        -------
        dydt : NDArray
            an array of time derivatives of A

        """
        dydt = np.zeros(y.shape)
        dydt[0] =(2*H*y[0] + k/a*y[1])
        dydt[1] = -k/a*y[0]
        return dydt

    
    def _find_tinit_bd(self, init : np.ndarray, mode : str="k") -> Tuple[np.ndarray, np.ndarray]:
        """
        Determines the pair of $k$ and $t$ satisfying $k = 10^(5/2)a(t)H(t)$.

        Depending on `mode`, `init` may be a time coordinate (`mode='t'`), $e$-folds (`mode='N'`) or momentum (`mode='k'`).

        Parameters
        ----------
        init : array
            the input array (t, N, or k)
        mode : str
            indicate the type of `init`

        Returns
        -------
        ks : NDarray
            an array of momenta
        tstart : NDarray
            an array of times

        Raises
        ------
        KeyError
            if `mode` is not 't, 'k' or 'N'
        """

        pwr = 5/2

        t = self._t
        def logkH(t): return np.log(self._af(t)*self._Hf(t))
        if mode=="t":
            tstart = init
            logks = logkH(tstart)
            ks = 10**(pwr)*np.exp(logks)

        elif mode=="k":
            ks = init

            tstart = []
            for k in ks:
                ttmp  = self._t[np.searchsorted(10**(pwr)*np.exp(logkH(self._t)), k, "right")]
                tstart.append(ttmp)
            tstart = np.array(tstart)

        elif mode=="N":
            tstart = CubicSpline(self._N, t)(init)
            ks = 10**pwr*np.exp(logkH(tstart))

        else:
            raise KeyError("'mode' must be 't', 'k' or 'N'")

        return ks, tstart
        
    def compute_ptvac(self, k : np.ndarray, uk : np.ndarray) -> np.ndarray:
        r"""
        Compute the vacuum power spectrum, $\mathcal{P}_{T,\lambda}^{\mathrm{vac}}(t,k)$.

        Parameters
        ----------
        k : NDArray
            momentum, $k$
        uk : NDArray
            vacuum tensor mode $u_0(t,k)$

        Return
        ------
        PTvac : NDArray
            the vacuum tensor power spectrum
        """
        PTvac = 2*(k*self._omega)**2/( np.pi**2 ) * abs(uk)**2
        return PTvac

    def compute_ptind(self, k : float, lgrav : float, Ngrid : np.ndarray, ind: int, GreenN : np.ndarray, pgrid : np.ndarray,
                                l1 : float, A1 : np.ndarray, dA1 : np.ndarray, l2 : float, A2 : np.ndarray, dA2 : np.ndarray,
                                momgrid : int=100) -> float:
        r"""
        Compute the vacuum power spectrum, $\mathcal{P}_{T,\lambda}^{\mathrm{vac}}(t,k)$.

        The integral is computed by integrating over the inner integral over the gauge-mode functions $A_\mu(s,p)$ using `scipy.integrate.trapezoid`.
        The external momentum integrals are also computed using `scipy.integrate.trapezoid` on a grid of momenta momgrid x momgrid. The mode functions
        are interpolated over momentum to match this grid.

        Parameters
        ----------
        k : float
            momentum $k$
        lgrav : NDArray
            helicity $\lambda$
        Ngrid : NDArray
            $e$-folds $\log a$
        ind : integer
            the power spectrum is computed at $t$ corresponding to `Ngrid[ind]`
        GreenN : NDArray
            the Green function $kG_k(t, s)$ with $s$ corresponding to `Ngrid`
        pgrid : NDArray
            momenta $p$ for the mode functions $A_\mu(s,p)$
        l1, l2 : float
            the gauge-field helicities $\mu_1$, $\mu_2$
        A1, A2 : float
            the gauge-field mode functions $\sqrt{2k} A_{\mu_1, \mu_2}(s, p)$
        dA1, dA2 : float
            the gauge-field mode function derivatives $a(s)\sqrt{2/k} \dot{A}_{\mu_1, \mu_2}(s, p)$
        momgrid : int
            the grid size for the momentum integrals.

        Returns
        -------
        PTind : float
            the gauge-field induced tensor power spectrum
        """

        cutUV = self._khN(Ngrid[ind])/k
        cutIR = min(pgrid)/k
        HN = self._HN(Ngrid)

        logAs = np.linspace(np.log(max(0.5, cutIR)), np.log(cutUV), momgrid)

        Ax_rad = PchipInterpolator(np.log(pgrid), abs(A1))
        Ax_phase = PchipInterpolator(np.log(pgrid), np.arccos(A1.real/abs(A1)))
        
        dAx_rad = PchipInterpolator(np.log(pgrid), abs(dA1))
        dAx_phase = PchipInterpolator(np.log(pgrid), np.arccos(dA1.real/abs(dA1)))

        Ay_rad = PchipInterpolator(np.log(pgrid), abs(A2))
        Ay_phase = PchipInterpolator(np.log(pgrid), np.arccos(A2.real/abs(A2)))
        
        dAy_rad = PchipInterpolator(np.log(pgrid), abs(dA2))
        dAy_phase = PchipInterpolator(np.log(pgrid), np.arccos(dA2.real/abs(dA2)))

        def Afuncx(x): return Ax_rad(x)*np.exp(1j*Ax_phase(x))
        def dAfuncx(x): return dAx_rad(x)*np.exp(1j*dAx_phase(x))

        def Afuncy(x): return Ay_rad(x)*np.exp(1j*Ay_phase(x))
        def dAfuncy(x): return dAy_rad(x)*np.exp(1j*dAy_phase(x))

        IntOuter = []
        for logA in logAs:
            A = np.exp(logA)
            Blow = (cutIR - A)
            Bhigh = (A - cutIR)
            if Bhigh>0.5:
                Blow = (A - cutUV)
                Bhigh = (cutUV - A)
                if Bhigh>0.5:
                    Blow = -0.5
                    Bhigh = 0.5
            Bs = np.linspace(Blow, Bhigh, momgrid)[1:-1]

            IntInner = np.zeros(Bs.shape)

            for j, B in enumerate(Bs):
                Ax = Afuncx(np.log(k*(A+B)))
                dAx = dAfuncx(np.log(k*(A+B)))
                Ay = Afuncy(np.log(k*(A-B)))
                dAy = dAfuncy(np.log(k*(A-B)))

                mom =  abs( l1*l2 + 2*lgrav*( (l1+l2)*A + (l1-l2)*B ) + 4*(A**2 - B**2) + 8*lgrav*A*B*( (l1-l2)*A - (l1+l2)*B ) - 16*l1*l2*A**2*B**2 )
                z = max(A+B,A-B)
                mask = np.where(z<self._khN(Ngrid)/k, 1, 0)

                val = (dAx*dAy + l1*l2*Ax*Ay)*mask*k/(np.exp(3*Ngrid)*HN)
                timeintegrand = GreenN*val*mom
            
                timeintre = trapezoid(timeintegrand[:ind].real, Ngrid[:ind])
                
                timeintim = trapezoid(timeintegrand[:ind].imag, Ngrid[:ind])
                
                IntInner[j] = (timeintre**2 + timeintim**2)*A

  
            IntOuter.append(trapezoid(IntInner, Bs))
        IntOuter = np.array(IntOuter)

        PTind = trapezoid(IntOuter, logAs) / (16*np.pi**4)*(k*self._omega)**4

        return PTind
    





r"""
This module defines constants and functions which are often times encountered in cosmology. These are primarily used in `GEFF.tools.gw`
"""
import natpy as nat
import numpy as np
import pandas as pd
import os

basepath = os.path.dirname(os.path.abspath(__file__))

nat.set_active_units("HEP")

#constants
G : float = 6.67430e-11 * nat.convert(nat.m**3 * nat.kg**(-1) * nat.s**(-2), nat.GeV**(-2)) 
"""Newtons constant in GeV."""
M_pl : float = (8*np.pi*G)**(-1/2) # reduced Planck mass in GeV
"""The reduced Planck mass in GeV."""

#Conversions
gev_to_hz : float = nat.convert(nat.GeV, nat.Hz)
"""The conversion factor between Hz and GeV."""

#Planck results
T_0 : float = 2.7255 * nat.convert(nat.K, nat.GeV)
"""The temperatur of CMB photons today (in GeV)."""
h : float = 0.674
"""The reduced Hubble parameter $h$ (from [Planck 2018](https://arxiv.org/abs/1807.06209))."""
H0 : float = h*100*nat.convert(nat.km * nat.s**(-1) * nat.Mpc**(-1), nat.GeV)
"""The Hubble rate today (in GeV)."""

gtab = pd.read_table(os.path.join(basepath, "../data/eff_dof.dat"), sep=" ", comment="#").values.T
gdic = {"T":gtab[0,:],"f":gtab[1,:],"grho":gtab[2,:], "gs":gtab[3,:]}

def g_rho(T : float) -> float:
    """
    Effective number of relativistic degrees of freedom as a function of temperature, $g_{*}(T)$.

    Data is taken from [2005.03544](https://arxiv.org/abs/2005.03544).
    """
    return np.interp(T, gdic["T"], gdic["grho"])

def g_s(T : float) -> float:
    """
    Effective number of entropic degrees of freedom as a function of temperature, $g_{*,S}(T)$.

    Data is taken from [2005.03544](https://arxiv.org/abs/2005.03544).
    """
    return np.interp(T, gdic["T"], gdic["gs"])

def g_rho_freq(f : float) -> float:
    """
    Effective number of relativistic degrees of freedom as a function of frequency, $g_{*}(T(f))$.

    Data is taken from [2005.03544](https://arxiv.org/abs/2005.03544).
    """
    return np.interp(f, gdic["f"], gdic["grho"])

def g_s_freq(f : float) -> float:
    """
    Effective number of entropic degrees of freedom as a function of frequency, $g_{*,S}(T(f))$.

    Data is taken from [2005.03544](https://arxiv.org/abs/2005.03544).
    """
    return np.interp(f, gdic["f"], gdic["gs"])



g_rho_0 : float = g_rho(T_0)
"""The effective number of relativistic degrees of freedom today, $g_{*}(T_0)$."""
g_s_0 : float = g_s(T_0)
"""The effective number of entropic degrees of freedom today, $g_{*,S}(T_0)$."""
omega_r : float = np.pi**2*g_rho_0/(90*M_pl**2*H0**2)*T_0**4
"""The density parameter for radiation today, $\Omega_r$."""

feq = 2.1e-17
"""Frequency corresponding to mattter-radiation equality"""




r"""
Defines the GEF model "classic" corresponding to pure axion inflation.

For more details on this model, see e.g., [arXiv:2109.01651](https://arxiv.org/abs/2109.01651).

---

The model knows the following variables:
* time variable: `t` - *cosmic time*, $t$ 
* dynamical variables:
    * `N` - *$e$-folds*,  $N$
    * `phi`, `dphi` - *inflaton amplitude, $\varphi$, and velocity, $\dot{\varphi}$* 
    * `kh` -  *the instability scale, $k_{\rm h}$*
* static variables:
    * `a` - *scale factor, $a$* 
    * `H` - *Hubble rate, $H$* 
    * `ddphi` - *inflaton acceleration, $\ddot{\varphi}$*
    * `E`, `B`, `G` - *gauge-field expectation values, $\langle {\bf E}^2 \rangle$, $\langle {\bf B}^2 \rangle$, -$\langle {\bf E} \cdot {\bf B} \rangle$*
    * `xi` - *instability parameter, $\xi$* 
* constants: 
    * `beta` - *coupling strength, $\beta$*
* functions: 
    * `V`,`dV` - *inflaton potential, $V(\varphi)$, and its derivative, $V_{,\varphi}(\varphi)$*
* gauge field: 
    * `GF` - *tower of gauge bilinears, $\mathcal{F}_{\mathcal X}^{(n)}$*

The model expects the following input:
* `phi`, `dphi` - *initial data on the inflaton, $\varphi$, $\dot\varphi$*
* `rhoChi` - *initial data on the fermion energy density, $\rho_{\chi}$*
* `beta` - *coupling strength, $\beta$*
* `V`, `dV` - *potential shape, $V(\varphi)$, $V_{,\varphi}(\varphi)$*

The model tracks the following events:
* end of inflation - terminate solver when $\ddot{a} < 0$
* negative norms - return an error when $\langle {\bf E}^2 \rangle$ or  $\langle {\bf B}^2 \rangle$ are negative 
"""
import numpy as np

from geff.bgtypes import t, N, a, H, phi, dphi, ddphi, V, dV, E, B, G, xi, kh, beta, GF
from geff.solver import TerminalEvent, ErrorEvent, GEFSolver
from geff.mbm import ModeSolver

from geff.utility.eom import friedmann, gauge_field_ode, dlnkh, klein_gordon, check_accelerated_expansion
from geff.utility.mode import bd_classic, mode_equation_classic
from geff.utility.boundary import boundary_pai
from geff.utility.general import heaviside
from geff._docs import generate_docs, docs_models


name : str = "PAI"
"""The models name."""

quantities : dict={
            "time":[t], #time coordinate according to which EoMs are expressed
            "dynamical":[N, phi, dphi, kh], #variables which evolve in time according to an EoM
            "static":[a, H, ddphi, xi, E, B, G], #variables which are derived from dynamical variables
            "constant":[beta], #constant quantities in the model
            "function":[V, dV], #functions of variables such as scalar potentials
            "gauge":[GF] #Gauge fields whose dynamics is given in terms of bilinear towers of expectation values
            }

#State which variables require input for initialisation
model_input = [ beta, phi, dphi, V, dV]

#this functions is called upon initialisation of the GEF class
def define_units(phi, dphi, V):
    #compute Hubble rate at t0
    rhoK = 0.5*dphi**2
    rhoV = V(phi)
    H0 = friedmann( rhoK, rhoV )
    
    omega = H0 #Characteristic frequency is the initial Hubble rate in Planck units
    mu = 1. #Charatcterisic amplitude is the Planck mass (in Planck units)

    return omega, mu

#define sys_to_yini for GEFSolver
def initial_conditions(sys, ntr):
    yini = np.zeros((ntr+1)*3+4)

    #from the 'input' dictionary
    yini[0] = sys.N.value
    yini[1] = sys.phi.value
    yini[2] = sys.dphi.value

    #needs to be computed
    sys.initialise("kh")( abs(sys.dphi)*sys.beta )

    #all gauge-field expectation values are assumed to be 0 at initialisation
    return yini

#define update_sys for GEFSolver
def update_values(t, y, sys):
    #spacetime variables
    sys.t.value = t
    sys.N.value = y[0]
    sys.a.value = np.exp(y[0])

    #parse for convenience
    sys.phi.value = y[1]
    sys.dphi.value =  y[2]
    sys.kh.value = np.exp(y[3])

    #the gauge-field terms in y are not stored, save these values here
    sys.E.value = y[4]*np.exp(4*(y[3]-y[0]))
    sys.B.value = y[5]*np.exp(4*(y[3]-y[0]))
    sys.G.value = y[6]*np.exp(4*(y[3]-y[0]))

    #Hubble rate
    sys.H.value = friedmann(0.5*sys.dphi**2, sys.V(sys.phi), 0.5*(sys.E+sys.B)*sys.omega**2)

    #boundary term parameter
    sys.xi.value = sys.beta*(sys.dphi/(2*sys.H))

    #acceleration for convenience
    sys.ddphi.value = klein_gordon(sys.dphi, sys.dV(sys.phi), sys.H, -sys.G*sys.beta*sys.omega**2)
    return

#define timestep for GEFSolver
def compute_timestep(t, y, sys):

    dydt = np.zeros(y.shape)

    #odes for N and phi
    dydt[0] = sys.H.value
    dydt[1] = sys.dphi.value
    dydt[2] = sys.ddphi.value

    #achieving dlnkhdt is monotonous requires some care
    dlnkhdt = dlnkh( sys.kh, sys.dphi, sys.ddphi, sys.beta,
                       0., sys.xi, sys.a, sys.H )
    
    logfc = y[0] + np.log( 2*abs(sys.xi)*dydt[0])
    dlnkhdt *= heaviside(dlnkhdt,0)*heaviside(logfc,y[3]*(1-1e-5))
    dydt[3] = dlnkhdt

    #compute boundary terms and then the gauge-field bilinear ODEs
    Fcol = y[4:].shape[0]//3
    F = y[4:].reshape(Fcol,3)
    W = boundary_pai(sys.xi.value)
    dFdt = gauge_field_ode(F, sys.a, sys.kh, 2*sys.H*sys.xi, W, dlnkhdt, L=20)

    #reshape to fit dydt
    dydt[4:] = dFdt.reshape(Fcol*3)

    return dydt

#Event 1: Track the end of inflation:
def condition_EndOfInflation(t, y, sys):
    #compute energy densities
    rhoV = sys.V(y[1])
    rhoK = y[2]**2/2
    rhoEM = 0.5*(y[4]+y[5])*(sys.omega/sys.mu)**2*np.exp(4*(y[3]-y[0]))

    #compute pressure
    pV = -rhoV
    pK = rhoK
    pEM = 1/3*rhoEM

    return check_accelerated_expansion([rhoV, rhoK, rhoEM], [pV, pK, pEM])/(sys.H)**2

def consequence_EndOfInflation(sys, occurrence):
    if occurrence:
        #stop solving once the end of inflation is reached
        return "finish", {}
    else:
        #increase tend given the current Hubble rate
        tdiff = np.round(5/sys.H, 0)
        #round again, sometimes floats cause problems in t_span and t_eval.
        tend  = np.round(sys.t + tdiff, 0)
        return "proceed", {"tend":tend}

EndOfInflation : TerminalEvent = TerminalEvent("End of inflation", condition_EndOfInflation, -1, consequence_EndOfInflation)
"""Defines the 'End of inflation' event."""

#Event 2: ensure that E^2 and B^2 are positive
def condition_NegativeNorms(t, y, sys):
    return min(y[4], y[5])
    
NegativeNorms : ErrorEvent = ErrorEvent("Negative norms", condition_NegativeNorms, -1, "Negative value for E^2 or B^2.")
"""Defines the 'Negative norms' event."""

events = [EndOfInflation, NegativeNorms]

#gather all information in the solver
solver = GEFSolver(initial_conditions, update_values, compute_timestep, quantities, events)
"""The solver used by the GEF model."""

#define mode-by-mode solver
MbM = ModeSolver(mode_equation_classic, {"a":a,"xi":xi, "H":H}, bd_classic, {})
"""The mode solver used by the GEF model."""


#define default docs for the above functions
generate_docs(docs_models.DOCS)








DOCS ={
    "module":r"""
        This module is designed to compute the tensor power spectrum from a GEF result.

        The polarized tensor power spectrum at time $t$ with momentum $k$ is given by
        $$\mathcal{P}_{T}(t, k) = \mathcal{P}_{T}^{\mathrm{vac}}(t, k) + \frac{1}{2}\sum_{\lambda=\pm 1} \mathcal{P}_{T,\lambda}^{\mathrm{ind}}(t, k)\, ,$$
        where $\lambda$ labels the helicity of tensor modes.

        The vacuum contribution is given by
        $$\mathcal{P}_{T,\lambda}^{\mathrm{vac}}(t, k) = \frac{4 k^3}{\pi^2 M_{\rm P}^2} |u_0(t, k)|^2\, ,$$
        while the induced contribution is
        $$\mathcal{P}_{T,\lambda}^{\mathrm{ind}}(t, k) = \frac{k^3}{2 \pi^2 M_{\rm P}^4} \int \frac{{\rm d}^3 {\bf p}}{(2 \pi)^3} \sum_{\alpha,\beta = \pm1} 
                \left(1 +  \lambda \alpha \frac{{\bf k} \cdot {\bf p}}{k p} \right)^2 \left(1 +  \lambda \beta \frac{k^2 - {\bf k} \cdot {\bf p}}{kq}  \right)^2 $$
        $$ \qquad \qquad \qquad \times \left|\int_{-\infty}^\infty {\rm d} s \frac{G_k(t, s)}{a^3(s)} 
            \left[A'_\alpha(s, p)A'_\beta(s, q) + \alpha \beta\, p q\, A_\alpha(s, p) A_\alpha(s, q) \right] \right|^2 \, ,
        $$
        with momentum $q = |{\bf p} + {\bf q}|$ and scale-factor $a$.

        The vacuum modes $u_0(t,k)$ obey the mode equation
        $$\mathcal{D}_k {u_0} = \ddot{u}_0 + 3 H \dot{u}_0 + \frac{k^2}{a^2} {u_0} = 0$$
        with the retarded Green function $G_k(t',t)$ defined for the operator $\mathcal{D}_k$.

        The gauge-field mode functions $A_\lambda(t,k)$ are defined as in the `geff.mode_by_mode` module.

        For details on the numerical computation, see the Appendix B of [2508.00798](https://arxiv.org/abs/2508.00798).
        """
}


DOCS = {
    "module":r"""
    This module is intended to compute a gravitational-wave spectrum from a tensor power spectrum.

    The gravitational-wave spectrum, $h^2\Omega_{\rm GW}$ is computed using `omega_gw`, which evaluates the formula

    $$\Omega_{\rm GW}(f) \equiv  \frac{1}{3 H_0^2 M_{\rm P}^2}\frac{{\rm d} \rho_{\rm GW} (f)}{{\rm d} \ln{f}} = 
                                    \frac{\pi^2}{3 H_0^2} f^2 |\mathcal{T}_{\rm GW}(f)|^2 \mathcal{P}_T(k_f), k_f) 
                                    \, , \quad k_f = 2 \pi a_0 f \, ,$$

    where  $\mathcal{P}_T(k)$ is the tensor power spectrum with momentum $k$ at the end of inflation, and $H_0$ is the Hubble rate today.
    The transfer function $|\mathcal{T}_{\rm GW}(f)|^2$ is given by

    $$|\mathcal{T}_{\rm GW}(f)|^2 \simeq \frac{H_0^2\Omega_r}{8 \pi^2 f^2} \frac{g_{\*}(T_f)}{g_{\*}(T_0)} 
                                            \left(\frac{g_{\*,S}(T_0)}{g_{\*,S}(T_f)}\right)^{4/3}
                                            \left(1 + \frac{9}{16}\left(\frac{f_{\rm eq}}{\sqrt{2} f} \right)^2\right)
                                            |\mathcal{T}_{\rm reh}(f)|^2 \,.$$

    It accounts for the evolution of $\Omega_{\rm GW}(f)$ from the end of inflation until today.
    Here, $T_f$ is the temperature corresponding to the frequency $f$.
    For $g_{\*}$, $g_{\*,S}$, $f_{\rm eq}$, $\Omega_r$, and $T_0$, we use the corresponding functions in `geff.utility.cosmo`.
    The term $|\mathcal{T}_{\rm reh}(f)|^2$ accounts for the transition through reheating. For instantaneous reheating, $|\mathcal{T}_{\rm reh}(f)|^2 = 1$.
    Otherwise, we assume that reheating proceeds via coherent oscillations of the inflaton field, such that

    $$
    |\mathcal{T}_{\rm reh}(f)|^2 = \frac{ \theta(f_{\rm end} - f)}{1 - 0.22 \left(\frac{f}{f_{\rm reh}}\right)^{1.5} + 0.65 \left(\frac{f}{f_{\rm reh}}\right)^{2}} \, ,
    $$

    as given in [arXiv:1407.4785](https://arxiv.org/abs/1407.4785) and [arXiv:2011.03323](https://arxiv.org/abs/2011.03323). 
    Here, $f_{\rm reh}$ and $f_{\rm end}$ are, respectively, the frequencies at the end of reheating, and inflation.

    The frequency $f$ can be computed from a comoving momentum $k$ using `k_to_f`, which evaluates

    $$
    f = \frac{k_f}{2 \pi a_0} =\frac{k_f}{2 \pi a_{\rm end}} e^{-N_{reh}} \left( \frac{g_{\*,S}(T_0)}{g_{\*,S}(T_{{\rm reh}})}\right)^{1/3} \frac{T_0}{T_{{\rm reh}}} \, .
    $$

    where

    $$
    N_{\rm reh} = \frac{1}{3(1 + w_{\rm reh})} \ln \left(\frac{90 M_{\rm P} H_{\rm end}^2}{\pi^2 g_*(T_{\rm reh}) T_{\rm reh}^4} \right) \, .
    $$
    and $T_{\rm reh}$ is the temperature of the SM plasma at the end of reheating.
    """
}


val_docs = """
    A `Quantity` subclass used as basis for defining real-valued variables and constants.
     
    In addition to the basic structure defined by `Quantity`, this class can be used like an arithmetic type.
    It defines basic arithmetic operations on its instances as operations on their underlying `value` attribute.

    As a subclass of `Quantity` it inherits all its attributes and methods, ensuring 
     that `value` is changed according to `units` by using `conversion`.

    The class serves as a parent for `Variable` and `Constant`. 
    """

variable_docs = """
    A `Val` subclass representing real-valued variables evolving with cosmic time.

    Instances of this class can be used like a `Val` with `value` as a 1-D Numpy-Array.
     Indexed returns the associated index of `value`.
    """

variable_addendum = r"""
    A typical `Variable` is the scalar field velocity, $\dot\varphi = \omega\mu \dot\bar{\varphi}$    

    To define a custom `Variable` object use the class factory `define_var`.
    """


bgvar_addendum = """
    This is a subclass of `Variable` with a custom name and scaling.
    """

constant_docs = """
    A `Val` subclass representing a constant of cosmic time.

    Instances of this class can be used like a float for mathematical operations as defined by `Val`.
    """

constant_addendum = r"""
    A typical `Constant` is the inflaton--gauge-field coupling, $\beta/M_{\rm P} \sim \beta/(\bar M_{\rm P} \mu)$    

    To define a custom `Constant` object use the class factory `define_const`.
    """


bgconst_addendum = """
    This is a subclass of `Constant` with a custom name and scaling.
    """


func_docs = """
    A `Quantity` subclass representing real functions of variables like the inflaton potential.
    
    An instance of this class can be used as a function,
    evaluating the underlying method, `basefunc` depending on the state of `units`.

    In physical units, the call returns the result of `basefunc`.
    In numerical units, the call instead returns `basefunc(*args)/conversion`.  
    If called by a `Val` object, the argument is also converted according to the units of the `Val` instance
    (generically, identical to the ones of the `Func` instance).
    If instead called by a regular arithmetic data type (e.g., `float`),
      it is assumed that the argument is in the same unit system as the `Func` instance.
    """
func_addendum = r"""
    A typical object is the scalar potential, $V(\varphi) = \omega^2 \mu^2 \bar{V}(\bar{\varphi} \mu) $

    To define a custom `Func` object, use the class factory `define_func`.
    """

bgfunc_addendum = """
    This is a subclass of `Func` with a custom name and scaling.
    """

gaugefield_docs = r"""
    A low level class defining some basic properties of gauge-field bilinear towers.

    A gauge-field bilinear tower is defined as a collection of the following three objects,

    $$ \mathcal{F}_\mathcal{E}^{(n)} = \frac{a^4}{k_{{\rm UV}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf E}\rangle = \int\limits_{0}^{k_{{\rm UV}}(t)}\frac{{\rm d} k}{k} \frac{a^2 k^{n+3}}{2 \pi^2 k_{{\rm UV}}^{n+4}}  \sum_{\lambda}\lambda^n |\dot{A}_\lambda(t,k)|^2\, ,$$
    $$ \mathcal{F}_\mathcal{G}^{(n)} = -\frac{a^4}{2 k_{{\rm UV}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf B} + {\bf B} \cdot \operatorname{rot}^n {\bf E}\rangle = \int\limits_{0}^{k_{{\rm UV}}(t)} \frac{{\rm d} k}{k} \frac{a k^{n+4}}{2 \pi^2 k_{{\rm UV}}^{n+4}}\sum_{\lambda}\lambda^{n+1} \operatorname{Re}[\dot{A}_\lambda(t,k)A_\lambda^*(t,k)] \, ,$$
    $$ \mathcal{F}_\mathcal{B}^{(n)} = \frac{a^4}{k_{{\rm UV}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf E}\rangle = \int\limits_{0}^{k_{{\rm UV}}(t)}\frac{{\rm d} k}{k} \frac{k^{n+5}}{2 \pi^{2}k_{{\rm UV}}^{n+4}} \sum_{\lambda}\lambda^n |A_\lambda(t,k)|^2 \, ,$$

    here $k_{\rm UV}$ is a UV cutoff scale, ${\bf E}$ and ${\bf B}$ are electric and magnetic field operators,
      and $A_\lambda(t,k)$ is a gauge-field mode function. The integer $n$ varies between $0$ and a maximum value, $n_{\rm tr}$.

    The `GaugeField` class collects important information about the collection $\mathcal{F}_\mathcal{X}^{(n)}$.
    The `GEFF` code needs to know, which `Variable` sets the UV cutoff scale, and
    which `Variable`s correspond to the zero-order quantities, $\langle {\bf E}^2\rangle$, $\langle {\bf B}^2\rangle$, and $-\langle {\bf E} \cdot {\bf B}\rangle$.

    Note that a `GaugeField` is never part of a `BGSystem`, as the number of variables depends on $n_{\rm tr}$, which is not fixed a priori.
    In fact, terms with $n>1$ are often only auxiliary and not used by the `GEFF` after the differential equations have been solved.
    However, the quantities $\mathcal{F}_\mathcal{X}^{(n)}$ are defined to be unitless, so they do not need to be converted between numerical and physical units.

    A custom `GaugeField` can be defined using `BGGauge`.
    """


DOCS = {
    "module":r"""
    This module defines base classes used throughout the `GEFF` module.

    The main purpose of these classes is to address the following.
    Most quantities appearing in the GEF will scale according to an inverse-time scale, $\omega$, and an energy scale, $\mu$.
    For example, a time derivative scales with inverse time, $\partial_t \sim \omega$. 
    Consequently, also a gauge field, $A_\mu \sim \omega$, as it appears in covariant derivatives.
    On the other hand, the amplitude of a scalar field scales with energy, $\varphi \sim \mu$.
    In typical inflationary contexts, $\omega \equiv H_0$, some constant Hubble rate, and $\mu \equiv M_{\rm P}$, the Planck mass.
    
    For numerical applications, it is convenient to work with dimensionless quantities.
    For example, it is useful to perform numerical computations using the dimensionless scalar amplitude $\bar{\varphi} = \varphi/\mu$, or the dimensionless gauge field, $\bar{A}_\mu = A_\mu / \omega$.
    Ultimately, the quantities we are interested in are obviously $\varphi$, $A_\mu$, etc. 
    Therefore, we need an easy way to switch between $\bar{X}$ and $X$.

    Throughout the code, we refer to the dimensionless variable, $\bar{X}$, as being in **numerical units**, while $X$ is in **physical units**. 

    To facilitate switching between these two unit systems throughout the code, this module provides the classes `BGSystem`, `Variable`, `Constant`, and `Func`. 
    The latter three are collectively referred to as `Quantity` objects.
    Each `Quantity` object is defined by a scaling with $\omega$ and $\mu$. For example, the `Quantity` $X$ scales as $X = \omega^a \mu^b \bar{X}$, where $\bar{X}$ is the re-scaled quantity used for numerical computations.
    A `BGSystem` is a collection of `Quantity` objects, which defines a common unit system by setting the value of $\omega$ and $\mu$.

    The user may define variables that evolve with cosmic time using the `define_var` class factory, which creates subclasses of `Variable`. Examples of a `Variable` are the Hubble rate, scalar field amplitude etc.
    In the same manner, the user can define constants of cosmic time using the `define_const` class factory, which creates subclasses of `Constant`. Examples of a `Constant` are, e.g., coupling strengths.
    Some quantities are functions of variables, for example, a scalar potential. These are defined by the factory `define_func`, which creates subclasses of `Func`.

    The following examples illustrates the basic use of these classes:

    Examples
    --------
    1. Defining a `BGSystem`
    ```python
    from GEFF.bgtypes import BGSystem, define_var

    # define two variables corresponding to physical time and Hubble rate.
    time = define_var("t", qu_omega=-1, qu_mu=0)
    Hubble = define_var("H", qu_omega=1, qu_mu=0)

    # Create a BGSystem with 'time' and 'Hubble'
    # We set the reference frequency to 1e-5*Mpl
    # The reference energy is the Planck mass in Planck units (so 1)
    U = BGSystem({time, Hubble}, omega=1e-5, mu=1)

    # The BGSystem knows about the new variables time and Hubble
    print(U.quantity_names()) # prints ["t", "H"]

    # However, neither the constant nor the variable are instantiated 
    # (we did not specify their value)
    print(U.variable_names()) # prints []  

    # Instantiate the quantities t and H using the units defined by U:
    U.initialise("t")(1e5)
    U.initialise("H")(1e-5)
    print(U.variable_names()) # prints ["t", "H"] 

    # The BGSystem now recognizes "t" and "H" as keys:
    print(f"In Planck units, the Hubble rate at time {U.t.value:.1e} is {U.H.value:.1e}.")
    # gives U.t=1e5, U.H=1e-5

    # Convert everything to numerical units
    U.units = False
    print(f"In numerical units, the Hubble rate at time {U.t.value} is {U.H.value:.1e}.") 
    # gives U.t=1, U.H=1

    ```

    2. Adding a new `Variable` 

    ```python
    #Let us reset the units of U:
    U.units = True

    # add a new Val: the electric-field expectation value E^2
    U.add_variable("E0", qu_omega=4, qu_mu=0) #since A_mu scales like d / dx^mu 

    # initialise E0 with some value in Planck units:
    U.initialise("E0")( 6e-10 )
    ```

    3. Operations between `Variable`'s

    ```python
    # We add a new Variable to U, the magnetic-field expectation value
    U.add_variable("B0", 4, 0)
    U.initialise("B0")(1e-10)

    # We can safely add up E0 and B0 as they are in the same units:
    print(U.E0 + U.B0) #gives 7e-10 = 6e-10 + 1e-10

    # This behaviour is not exclusive to Variables,
    # but also works for Constants:
    U.add_constant("BConst", 4, 0)
    U.initialise("BConst")(5e-11)

    print(U.E0 + U.BConst) #gives 6.5e-10 = 6e-10 + 5e-11
    ```

    4. Changing the value of a `Variable`
    ```python
    # The value of a `Variable` can be passed directly
    U.BConst.value = 1e-3
    print(str(U.BConst))

    # if we change to numerical units,
    # the argument which is passed is treated in numerical units
    U.units = False
    U.BConst.value = 1
    print(U.BConst)
    ```

    5. Adding a new `Func`

    ```python
    #first, return to physical units
    U.units = True

    # define a new Func: rhoE, the electric-field energy density
    rhoE = U.add_function("rhoE", args=[U.E0], qu_omega=2, qu_mu=2) # since 3 * M_pl^2 * H^2 = rho
    #Note how E0 is passed for creation to indicate the scaling of the argument

    # define rhoE as a function of E0:
    def func(x): return 0.5*x
    U.initialise("rhoE")( func )
    # U.rhoE is now a Callable function with a single argument
    ```

    6. Calling a `Func` with a `Val`

    ```python
    # Calling rhoE in physical units is straight forwards:
    print( U.rhoE( U.E0 ) ) # gives 3e-10 (in Planck units)

    # compare this to a direct call to func:
    print( func(6e-10) )  # gives 3e-10 (the result in Planck units)

    # Switching E0 to numerical units, nothing changes since rhoE is in physical units:
    U.E0.units = False # E0.value = 6e10
    print( U.rhoE(U.E0) ) # gives 3e-10 (still in Planck units)

    # Only switching U.rhoE to numerical units changes the result:
    U.rhoE.units = False
    print( U.rhoE(U.E0) ) # gives 3. = 3e-10 / (U.omega*U.mu)**2 (now in numerical units)

    # Again, this outcome does not depend on the units of E0:
    U.E0.units = False
    print( U.rhoE(U.E0) ) # gives 3. = 3e-10 / (U.omega*U.mu)**2 (in numerical units)
    ```

    7. Calling a `Func` with a `float`

    ```python
    # instead of calling rhoE by E0, we can call it by a float:
    val = 6e-10

    # First the behavior if rhoE is in physical units:
    U.rhoE.units = True
    print( U.rhoE( val ) ) # gives 3e-10 (in Planck units)

    # The behavior is different compared to a Val if rhoE is in numerical units:
    U.rhoE.units = False
    print( U.rhoE(val) ) #gives 3e-20 = 0.5* (6e-10*U.omega**4) / (U.omega*U.mu)**2

    # Unlike a Val, the float does not track units
    # Therefore, it is always assumed to be in the units of rhoE
    # If you want to get the correct result, you would need to convert val by hand:
    print( U.rhoE(val/U.omega**4) ) # gives 3., the expected result in numerical units.

    # Overall, its safer to just keep everything in the same units:
    U.units = True
    ```
    """,

    "BGSystem":"""
    A collection of cosmological variables sharing a system of units.

    Instances of this class define two base unit systems,
    *physical units* and *numerical units*, by setting an energy scale, `mu`, and an inverse time scale, `omega`. 

    The cosmological variables (time, Hubble rate, etc.) are represented by `Quantity` objects.
    These objects are stored in `quantities`, and can can be initialise using `initialise`.
    Instances of these objects can be collectively converted between units by setting `units`. 

    This class is the fundamental building block of the `GEFF` code. 
    """,

    "Quantity":r"""
    An object representing a cosmological quantity. 

    A cosmological quantity has a characteristic scaling with respect to a frequency scale (e.g., the Hubble rate at some reference time), and energy scale (e.g., the Planck mass).

    Typical such objects are:
    - cosmic time $t = \bar{t}/\omega$
    - frequency $f = \omega \bar{f}$
    - scalar-field vev $\varphi = \mu \bar{\varphi}$
    - scalar potential $ V(\varphi) = \omega^2 \mu^2 \bar{V}(\bar{\varphi} \mu) $ 
    - gauge fields $A_\mu =  \omega\bar{A}_\mu $ *(i.e., scales like a time derivative)*

    `Quantity` objects are initialized as part of a `BGSystem`, which defines $\omega$ and $\mu$.

    This class is a parent of `Val` and `Func`.
    """,

    "Val":val_docs,
    "Variable":variable_docs,
    "Constant":constant_docs,
    
    "define_var.CustomVar":variable_docs+bgvar_addendum,
    "define_const.CustomConst":variable_docs+bgvar_addendum,

    "Func":func_docs+ func_addendum,
    "define_func.CustomFunc":val_docs+bgfunc_addendum,

    "GaugeField":gaugefield_docs,
}




r"""
Module defining the GEF model "SE no-scale" corresponding to fermionic axion inflation without a heuristic scale dependence.

For more details on this model, see e.g., [2109.01651](https://arxiv.org/abs/2109.01651).

---

The model knows the following variables:
* time variable: `t` - *cosmic time*, $t$ 
* dynamical variables:
    * `N` - *$e$-folds*,  $N$
    * `phi`, `dphi` - *inflaton amplitude, $\varphi$, and velocity, $\dot{\varphi}$* 
    * `kh` -  *the instability scale, $k_{\rm h}$*
    * `delta` - *cumulative electric damping, $\Delta$*
    * `rhoChi` - *fermion energy density, $\rho_{\chi}$*
* static variables:
    * `a` - *scale factor, $a$* 
    * `H` - *Hubble rate, $H$* 
    * `ddphi` - *inflaton acceleration, $\ddot{\varphi}$*
    * `E`, `B`, `G` - *gauge-field expectation values, $\langle {\bf E}^2 \rangle$, $\langle {\bf B}^2 \rangle$, -$\langle {\bf E} \cdot {\bf B} \rangle$*
    * `xi` - *instability parameter, $\xi$* 
    * `sigmaE`, `sigmaB` - *electric and magnetic conductivities, $\sigma_{\rm E}$, $\sigma_{\rm B}$*
    * `xieff` - *effective instability parameter, $\xi_{\rm eff}$*
    * `s` - *electric damping parameter, $s = \sigma_{\rm E}/(2H)$*
* constants: 
    * `beta` - *coupling strength, $\beta$*
* functions: 
    * `V`,`dV` - *inflaton potential, $V(\varphi)$, and its derivative, $V_{,\varphi}(\varphi)$*
* gauge field: 
    * `GF` - *tower of gauge bilinears, $\mathcal{F}_{\mathcal X}^{(n)}$*

The model expects the following input:
* `phi`, `dphi` - *initial data on the inflaton, $\varphi$, $\dot\varphi$*
* `rhoChi` - *initial data on the fermion energy density, $\rho_{\chi}$*
* `beta` - *coupling strength, $\beta$*
* `V`, `dV` - *potential shape, $V(\varphi)$, $V_{,\varphi}(\varphi)$*

The model tracks the following events:
* end of inflation - terminate solver when $\ddot{a} < 0$
* negative norms - return an error when $\langle {\bf E}^2 \rangle$ or  $\langle {\bf B}^2 \rangle$ are negative 
"""
import numpy as np

from geff.bgtypes import t, N, a, H, phi, dphi, ddphi, V, dV, E, B, G, xi, kh, beta, GF, define_var
from geff.solver import TerminalEvent, ErrorEvent, GEFSolver
from geff.mbm import ModeSolver

from geff.utility.eom import (klein_gordon, friedmann, dlnkh_schwinger, ddelta, drhoChi, gauge_field_ode_schwinger,
                                        conductivities_collinear, conductivities_mixed, check_accelerated_expansion)
from geff.utility.boundary import boundary_fai
from geff.utility.general import heaviside
from geff.utility.mode import mode_equation_SE_no_scale, damped_bd
from geff._docs import generate_docs, docs_models

name = "FAI basic"
"""The models name."""

settings = {"picture":"mixed"}
"""The model settings.

Possible settings are "mixed", "electric", "magnetic".

Determines if conductivities are computed assuming collinear E&M fields 
("electric", "magnetic") or not ("mixed").
"""

# parse settings
def define_conductivity():
    if settings["picture"]=="mixed":
        conductivity = conductivities_mixed
    elif settings["picture"]=="electric":
        def conductivity(a, H, E, B, G, omega):
            return conductivities_collinear(a, H, E, B, G, -1, omega)
    elif settings["picture"]=="magnetic":
        def conductivity(a, H, E, B, G, omega):
            return conductivities_collinear(a, H, E, B, G, 1, omega)
    else:
        raise KeyError(f"{settings['picture']} is an unknown choice for the setting 'picture'")
    return np.vectorize(conductivity)

def interpret_settings():
    global conductivity
    conductivity = define_conductivity()
    return

#Define all additional variables
sigmaE=define_var("sigmaE", 1, 0, "electric damping")
sigmaB=define_var("sigmaB", 1, 0, "magnetic damping")
delta=define_var("delta", 0, 0, "cumulative electric damping") 
xieff=define_var("xieff", 0, 0, "effective instabilty parameter") 
s=define_var("s", 0, 0, "electric damping parameter")
rhoChi=define_var("rhoChi", 4, 0, "fermion energy density")

#Assign quantities to a dictionary, classifying them by their role:
quantities={
            "time":[t],
            "dynamical":[N, phi, dphi, kh, delta, rhoChi],
            "static":[a, H, xi, E, B, G, ddphi, sigmaE, sigmaB, xieff, s],
            "constant":[beta],
            "function":[V, dV],
            "gauge":[GF]
            }

#State which variables require input for initialisation
model_input = [phi, dphi, rhoChi, beta, V, dV]

#this functions is called upon initialisation of the GEF class
def define_units(phi, dphi, V, rhoChi):
    #compute Hubble rate at t0
    rhoK = 0.5*dphi**2
    rhoV = V(phi)
    H0 = friedmann( rhoK, rhoV, rhoChi )
    
    omega = H0 #Characteristic frequency is the initial Hubble rate
    mu = 1. #Charatcterisic amplitude is the Planck mass
    
    return omega, mu

#the new function for sys_to_yini in GEFSolver
def initial_conditions(sys, ntr):
    yini = np.zeros((ntr+1)*3+6)

    #from the 'input' dictionary
    yini[0] = sys.N.value
    yini[1] = sys.phi.value
    yini[2] = sys.dphi.value

    #needs to be computed
    sys.initialise("kh")( abs(sys.dphi)*sys.beta )
    yini[3] = np.log(sys.kh.value)

    #initialise delta and rhoChi
    yini[4] = sys.delta.value
    yini[5] = sys.rhoChi.value

    #all gauge-field expectation values are assumed to be 0 at initialisation
    return yini

#define update_sys for GEFSolver
def update_values(t, y, sys, atol=1e-20, rtol=1e-6):
    #spacetime variables
    sys.t.value = t
    sys.N.value = y[0]
    sys.a.value = np.exp(y[0])

    #parse for convenience
    sys.phi.value = y[1]
    sys.dphi.value = y[2]
    sys.kh.value = np.exp(y[3])
    sys.delta.value = y[4]
    sys.rhoChi.value = y[5]

    #the gauge-field terms in y are not stored, save these values here
    sys.E.value = y[6]*np.exp(4*(y[3]-y[0]))
    sys.B.value = y[7]*np.exp(4*(y[3]-y[0]))
    sys.G.value = y[8]*np.exp(4*(y[3]-y[0]))

    #Hubble rate
    sys.H.value = ( friedmann(0.5*sys.dphi**2, sys.V(sys.phi), 
                                0.5*(sys.E+sys.B)*sys.omega**2, sys.rhoChi*sys.omega**2) )

    #conductivities
    sigmaE, sigmaB, ks = conductivity(sys.a.value, sys.H.value, sys.E.value,
                                       sys.B.value, sys.G.value, sys.omega)

    GlobalFerm = heaviside(np.log(ks),np.log(sys.a*sys.H))
    sys.sigmaE.value = (GlobalFerm*sigmaE)
    sys.sigmaB.value = (GlobalFerm*sigmaB)

    #boundary term parameters
    sys.s.value = sys.sigmaE/(2*sys.H)
    sys.xi.value = sys.beta*(sys.dphi/(2*sys.H))
    sys.xieff.value = sys.xi + sys.sigmaB/(2*sys.H)

    #acceleration for convenience
    sys.ddphi.value = klein_gordon(sys.dphi, sys.dV(sys.phi), sys.H, -sys.G*sys.beta*sys.omega**2)
    return

def compute_timestep(t, y, sys, atol=1e-20, rtol=1e-6):
    dydt = np.zeros(y.shape)

    #odes for N and phi
    dydt[0] = sys.H.value
    dydt[1] = sys.dphi.value
    dydt[2] = sys.ddphi.value

    #achieving dlnkhdt is monotonous requires some care
    dlnkhdt = dlnkh_schwinger( sys.kh, sys.dphi, sys.ddphi, sys.beta,
                                        0., sys.xieff, sys.s, sys.a, sys.H )
    xieff = sys.xieff.value
    s = sys.s.value
    sqrtterm = np.sqrt(xieff**2 + s**2 + s)
    r = (abs(xieff) + sqrtterm)
    logfc = y[0] + np.log(r*dydt[0]) 
    dlnkhdt *= heaviside(dlnkhdt,0)*heaviside(logfc,y[3]*(1-1e-5))
    dydt[3] = dlnkhdt

    #the other derivatives are straight forwards
    dydt[4] = ddelta( sys.delta, sys.sigmaE )
    dydt[5] = drhoChi( sys.rhoChi, sys.E, sys.G, sys.sigmaE, sys.sigmaB, sys.H )

    #compute boundary terms and then the gauge-field bilinear ODEs
    Fcol = y[6:].shape[0]//3
    F = y[6:].reshape(Fcol,3)
    W = boundary_fai(sys.xieff.value, sys.s.value)
    dFdt = gauge_field_ode_schwinger( F, sys.a, sys.kh, 2*sys.H*sys.xieff,
                                            sys.sigmaE, sys.delta,
                                                W, dlnkhdt )
    #reshape to fit dydt
    dydt[6:] = dFdt.reshape(Fcol*3)

    return dydt

#Event 1: Track the end of inflation:
def condition_EndOfInflation(t, y, sys):
   #compute energy densities
    rhoV = sys.V(y[1])
    rhoK = y[2]**2/2
    rhoEM = 0.5*(y[6]+y[7])*(sys.omega/sys.mu)**2*np.exp(4*(y[3]-y[0]))
    rhoF = y[5]*(sys.omega/sys.mu)**2

    #compute pressure
    pV = -rhoV
    pK = rhoK
    pEM = 1/3*rhoEM
    pF = 1/3*rhoF

    return check_accelerated_expansion([rhoV, rhoK, rhoEM, rhoF], [pV, pK, pEM, pF])/(sys.H)**2

def consequence_EndOfInflation(sys, occurance):
    if occurance:
        #stop solving once the end of inflation is reached
        return "finish", {}
    else:
        #increase tend given the current Hubble rate
        tdiff = np.round(5/sys.H, 0)
        #round again, sometimes floats cause problems in t_span and t_eval.
        tend  = np.round(sys.t + tdiff, 0)
        return "proceed", {"tend":tend}
    
EndOfInflation = TerminalEvent("End of inflation", condition_EndOfInflation, -1, consequence_EndOfInflation)
"""Defines the 'End of inflation' event."""

#Event 2: ensure that E^2 and B^2 are positive
def condition_NegativeNorms(t, y, sys):
    return min(y[6], y[7])
    
NegativeNorms : ErrorEvent = ErrorEvent("Negative norms", condition_NegativeNorms, -1, "Negative value for E^2 or B^2.")
"""Defines the 'Negative norms' event."""

events = [EndOfInflation, NegativeNorms]



#gather all information in the solver
solver = GEFSolver(initial_conditions, update_values, compute_timestep, quantities, events)
"""The solver used by the GEF model."""

#define mode-by-mode solver
MbM = ModeSolver(mode_equation_SE_no_scale, {"a":a, "xieff":xieff, "H":H, "sigmaE":sigmaE},
                         damped_bd, {"a":a, "sigmaE":sigmaE, "delta":delta}, default_atol=1e-5)
"""The mode solver used by the GEF model."""


#define default docs for the above functions
generate_docs(docs_models.DOCS)



import numpy as np
from .bgtypes import BGSystem, t, N, a, H
from .mbm import SpecSlice
from scipy.integrate import solve_ivp
from typing import Callable, Tuple, ClassVar
from ._docs import generate_docs, docs_solver
from .utility.general import AuxTol

class BaseGEFSolver:
    known_variables : ClassVar[dict] = {"time":[t], "dynamical":[N], "static":[a], "constant":[H], "function":[], "gauge":[]}
    """
    Classifies variables used by the solver according to:
    * 'time': the name of the time parameter of the ODE's (should be "t") 
    * 'dynamical': variables evolved by the EoM (not 'gauge')
    * 'gauge': tower of gauge-field expectation values evolved by the EoM
    * 'static': variables computed from 'dynamical' and 'gauge'
    * 'constant': constant variables
    * 'function': functions of the above variables.
    """

    known_events : ClassVar[dict] = {}
    """The `Event` objects which are tracked by the solver."""

    def __init__(self, init_sys : BGSystem):
        """
        Pass initial data to the solver.

        Parameters
        ----------
        init_sys : BGSystem
            initial data used by the solver
        """

        self.init_vals : BGSystem = BGSystem.from_system(init_sys, copy=True)
        """Initial data for the EoM's defined at $t_0 = 0$."""

        self.init_vals.units = False

        self.settings : dict ={"atol":1e-20, "rtol":1e-6, "attempts":5, "solvermethod":"RK45", "ntrstep":10}
        """
        A dictionary of internal settings used by the class:
        - atol: absolute tolerance used by `solve_ivp` (default: 1e-20)
        - rtol: relative tolerance used by `solve_ivp` (default: 1e-6)
        - solvermethod: method used by `solve_ivp` (default: 'RK45')
        - attempts: attempts made by `compute_GEF_solution` (default: 5)
        - ntrstep: `ntr` increment used in `compute_GEF_solution` (default: 10)
        """

        self.ntr : int = 100
        r"""Truncation number $n_{\rm tr}$, for truncated towers of ODE's."""

        self.tend : float = 120.
        r"""The time $t_{\rm end}$ up to which the EoMs are solved."""

        #initial conditions on initialisation
        self.set_initial_conditions_to_default()

        self._test_ode()

    @classmethod
    def toggle_event(self, event_name : str, toggle : bool):
        """
        Disable or enable an `Event` for the solver.

        Parameters
        ----------
        event_name : str
            the name of the target
        toggle : bool
            if the event should be active or inactive
        """
        if event_name in [event for event in self.known_events.keys()]:
            self.known_events[event_name].active = toggle
            humanreadable = {True:"active", False:"inactive"}
            print(f"The event '{event_name}' is now {humanreadable[toggle]}")
        else:
            print(f"Unknown event: '{event_name}'")
        return

    def compute_GEF_solution(self):
        """
        An algorithm to solve the GEF equations.

        The solver attempts to solve the EoMs several times using `solve_eom`. 
        If this returns a `TruncationError`, `ntr` is increased by `settings['ntrstep']`.
        Afterwards, `solve_eom` is called again until it returns a result.
        This is done for `settings['attempts']` times or until `ntr=200` is reached.

        Returns
        -------
        sol
            a bunch object returned by `solve_ivp` containing the solution

        Raises
        ------
        RuntimeError
             if no solution was obtained after the maximum number of attempts.
        """
        maxntr = 200
        maxattempts = self.settings["attempts"]
        ntrstep = self.settings["ntrstep"]

        attempt=0
        done = False
        sol = None
        #Run GEF
        while not(done) and (attempt<maxattempts):
            attempt+=1
            try:
                #Try to get convergent solution
                t0, yini, vals = self.initial_conditions()
                sol = self.solve_eom(t0, yini, vals)
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except TruncationError:
                if self.ntr<maxntr:
                    self._increase_ntr( min(ntrstep, maxntr-self.ntr) )
                else:
                    print("Cannot increase ntr further.")
                    break
            else:
                done=True
        
        if sol is None:
            raise RuntimeError(f"No GEF solution after {attempt} attempts.")
        else:
            if sol.success:
                print("Successful GEF solution obtained. Proceeding.\n")
            else:
                print("Processing unsuccessful solution.\n")
        
        return sol
    
    ### handle initial conditions ###

    @staticmethod
    def vals_to_yini(vals : BGSystem, ntr : int) -> np.ndarray:
        """
        Create an array of initial data from a BGSystem.

        Parameters
        ----------
        vals : BGSystem
            a unit system with initial data
        ntr : int
            truncation number (relevant for dynamical gauge-fields)

        Returns
        -------
        yini : NDarray
            an array of initial data
        """
        vals.units = False #ensure the system is in numerical units

        #In the simple version, ntr has no meaning, as there are no gauge fields
        yini = np.zeros((1)) 
        #get N from vals
        yini[0] = vals.N.value 
        return yini

    def set_initial_conditions_to_default(self):
        """
        Configure the solver to use initial data from `init_vals` using `vals_to_yini`.
        """
        def default_initial_conditions():
            """
            Compute initial data with `init_vals` using `vals_to_yini`.
            """
            t0 = 0
            vals = BGSystem.from_system(self.init_vals, True)
            vals.units = False
            yini = self.vals_to_yini(vals, self.ntr)
            return t0, yini, vals
        self.initial_conditions = staticmethod(default_initial_conditions)
        return
        
    
    def set_initial_conditions_to_MbM(self, sol, reinit_spec : SpecSlice):
        r"""
        Configure the solver to initialise data from a mode-by-mode solution.

        Used for mode-by-mode self correction. 
        Gauge-bilinears, $F_{\mathcal X}^{(n>1)}$ are re-initialized using `.mbm.SpecSlice.integrate_slice`.
        
        Parameters
        ----------
        sol
            a bunch object returned by `solve_ivp` containing the solution
        reinit_spec : SpecSlice
            spectrum at time of reinitialization
        """
        def MbM_initial_conditions():
            ntr = self.ntr

            t_reinit = reinit_spec["t"]

            reinit_ind = np.searchsorted(sol.t,t_reinit, "left")

            #Create unit system (copy to also get constants and functions):
            temp = BGSystem.from_system(self.init_vals, copy=True)

            #get number of regular dynamical variables
            n_dynam = len(self.known_variables["dynamical"])

            #construct yini
            yini = np.zeros(((ntr+1)*3+n_dynam))

            #parse everyting except for GEF-bilinears with n>1 to yini
            yini[:n_dynam+3] = sol.y[:n_dynam+3,reinit_ind]

            # compute En, Bn, Gn, for n>1 from Modes
            yini[n_dynam+3:] = np.array([reinit_spec.integrate_slice(n=n, integrator="simpson")
                                            for n in range(1,ntr+1)])[:,:,0].reshape(3*(ntr))
            
            self.parse_arr_to_sys(t_reinit, yini, temp)

            return t_reinit, yini, temp
        self.initial_conditions = staticmethod(MbM_initial_conditions)
        return
    
    ### Define ODE ###
    
    @staticmethod
    def update_vals(t : float, y : np.ndarray, vals : BGSystem):
        """
        Translate an array of data at time t into a `BGSystem`.

        Parameters
        ----------
        t : float
            time coordinate
        y : NDArray:
            array of data at time t
        vals : BGSystem
            the target system
        atol : float
            absolute tolerance parameters (if needed)
        rtol : float
            relative tolerance parameters (if needed)
        """

        #parse dynamical variables back to vals:
        vals.t.value = t
        vals.N.value = y[0]

        #compute static variables from y:
        vals.a.value = (np.exp(y[0]))

        #in case heaviside functions are needed in update_vals, atol and rtol are passed
        return
    
    @staticmethod
    def timestep(t : float, y : np.ndarray, vals : BGSystem) -> np.ndarray:
        """
        Compute time derivatives for data at time t using a `BGSystem`.

        Parameters
        ----------
        t : float
            a time coordinate
        y : NDArray:
            an array of data at time t
        vals : BGSystem
            the system used to compute the derivative
        atol : float
            absolute tolerance parameters (if needed)
        rtol : float
            relative tolerance parameters (if needed)

        Returns
        -------
        dydt : NDArray
            the time derivative of y
        """
        dydt = np.zeros_like(y)

        dydt[0] = vals.H.value #use the constant hubble rate to evolve N
        return dydt
    
    def ode(self, t : float, y : np.ndarray, vals : BGSystem) -> np.ndarray:
        """
        subsequently call `update_vals` and `timestep` to formulate an ODE for `solve_ivp`.

        Parameters
        ----------
        t : float
            a time coordinate
        y : NDArray:
            an array of data at time t
        vals : BGSystem
            the system passed to `update_vals` and `timestep` and 

        Returns
        -------
        dydt : NDArray
            the time derivative of y
        """
        self.update_vals(t, y, vals)
        dydt = self.timestep(t, y, vals)
        return dydt
    
    ### solve EoMs ###
    
    def solve_eom(self, t0 : float, yini : np.ndarray, vals : BGSystem):
        """
        Attempt to solve the GEF EoM's using `scipy.integrate.solve_ivp`.

        The solver attempts to obtain a GEF solution. This solution is then checked for any `ErrorEvent` occurrences.
        In this case, the solver marks the solution as unsuccessful and returns it for further processing.
        If no `ErrorEvent` occurrences are found, the other `TerminalEvent` occurrences are analyzed.
        These decide if the solution is returned, repeated, or if the solver should continue to solve.
        If there are no active `TerminalEvent` instances, the solution is returned after reaching `settings['tend']`.

        Parameters
        ----------
        t0 : float
            time of initialization
        yini : NDArray
            initial data at t0
        vals : BGSystem
            evolved alongside yini

        Returns
        -------
        sol
            a bunch object returned by `solve_ivp` containing the solution

        Raises
        ------
        TruncationError: 
            if an internal error occurred while solving the ODE.
        RuntimeError:
            if no 'finish' or 'error' command is obtained from an `Event` and `settings['tend']` is also not reached.
        """
        done = False
        attempts = 0
        sols = []

        event_dict, event_funcs = self._setup_events()

        print(f"The GEFSolver aims at reaching t={self.tend} with ntr={self.ntr}.")
        while not(done) and attempts < 10:

            try:
                tend = self.tend
                atol = self.settings["atol"]
                rtol = self.settings["rtol"]
                solvermethod = self.settings["solvermethod"]

                teval = np.arange(np.ceil(10*t0), np.floor(10*tend) +1)/10 #hotfix

                sol = solve_ivp(self.ode, [t0,tend], yini, t_eval=teval, args=(vals,),
                                method=solvermethod, atol=atol, rtol=rtol, events=event_funcs)
                if not(sol.success):
                    raise TruncationError
            #find 
            except KeyboardInterrupt:
                raise KeyboardInterrupt(f"Interrupted solver at t={float(vals.t.value):.2f}, N={float(vals.N.value):.2f}.")
            except Exception as e:
                print(f"While solving the GEF ODE, an error occurred at t={float(vals.t.value):.2f}, N={float(vals.N.value):.2f}):")
                print(e)
                raise TruncationError
            
            else:
                sols.append(sol)

                event_dict_new, command, terminal_event = self._assess_event_occurrences(sol.t_events, sol.y_events, vals)

                for key in event_dict_new.keys():
                    event_dict[key].append(event_dict_new[key])

                if command in ["finish", "error"]:
                    done=True
                elif command=="repeat":
                    print("Repeating")
                    sols.pop()
                elif command=="proceed":
                    #print("Proceeding")
                    t0 = sol.t[-1]
                    yini = sol.y[:,-1]

        sol = self._finalise_solution(sols, event_dict, terminal_event)
    
        if attempts != 1 and not(done):
            print(f"The run failed after {attempts} attempts.")
            raise RuntimeError
        
        return sol

    
    def _setup_events(self):
        event_funcs = []
        event_dict = {}
        for name, event in self.known_events.items():
            if event.active:
                event_funcs.append(event.event_func)
                event_dict[name] = []
        
        return event_dict, event_funcs
    
    
    def _assess_event_occurrences(self, t_events, y_events, vals):
        
        event_dict = {}

        active_events = [event for event in self.known_events.values() if event.active]

        # record which events have occurred
        occurrences = {}
        for i, event in enumerate(active_events):

            #Check if the event occured
            occurrence = (len(t_events[i]) != 0)

            #Add the event occurrence to the event dictionary:
            if occurrence:
                event_dict.update({event.name:t_events[i]})
                if len(t_events[i])>1:
                    tstr = [f"{t:.1f}" for t in t_events[i]]
                else:
                    tstr = f"{t_events[i][0]:.1f}"
                print(f"{event.name} at t={tstr}")
            occurrences[event.name] = occurrence

        error_events = [event for event in active_events if event.type=="error"]
        
        
        # Treat occurrences of ErrorEvents 
        for event in error_events:
            if occurrences[event.name]:
                print(f"Error: {event.message}")
                return event_dict, "error", event.name
                
        # if not ErrorEvents occurred, treat TerminalEvents
        terminal_events = [event for event in active_events if event.type=="terminal"]
        commands = {"primary":[], "secondary":[]}
        
        for event in terminal_events:
            #Asses the events consequences based on its occurrence or non-occurrence
            primary, secondary = event.event_consequence(vals, occurrences[event.name])
                
            for key, item in {"primary":(primary, event.name), "secondary":secondary}.items(): 
                commands[key].append(item)

        # If no error has occurred, handle secondary commands
        for command in commands["secondary"]:
            for key, item in command.items():
                if key =="timestep":
                    setattr(self, key, item)
                elif key=="tend":
                    if item > self.tend:
                        print(f"Increasing tend by {item-self.tend:.1f} to {item:.1f}")
                        setattr(self, key, item)
                elif key in self.settings.keys():
                    self.update_settings({key:item})
                else:
                    print(f"Unknown setting '{key}', ignoring input.")
        
        #Check command priority (in case of multiple terminal events). finish > repeat > proceed
        for primary_command in ["finish", "repeat", "proceed"]:
            for item in commands["primary"]:
                command = item[0]
                trigger = item[1]
                if command == primary_command:
                    return event_dict, command, trigger 

        #if no primarycommand was passed, return "finish" (no TerminalEvent or ErrorEvents)
        return event_dict, "finish", None
    
    def _finalise_solution(self, sols, event_dict, trigger_event):
        nfevs = 0
        y = []
        t = []
        solution = sols[-1]
        for s in sols:
            t.append(s.t[:-1])
            y.append(s.y[:,:-1])
            nfevs += s.nfev
        t.append(np.array([s.t[-1]]))
        y.append(np.array([s.y[:,-1]]).T)

        t = np.concatenate(t)
        y = np.concatenate(y, axis=1)

        solution.t = t
        solution.y = y
        solution.nfev = nfevs

        if trigger_event is None:
            solution.success = True
        else:
            solution.success = (trigger_event not in [event.name for event in self.known_events.values() if event.active and event.type=="error"])
            solution.message = f"A terminal event occured: '{trigger_event}'"

        for event_name in (event_dict.keys()):
            try:
                event_dict[event_name] = np.concatenate(event_dict[event_name])
            except ValueError:
                event_dict[event_name] = np.array(event_dict[event_name])

        solution.events = event_dict
        return solution
    
    ### Utility fumctions ###
    
    #can become an internal-only method once the GEFSolver output is changed.
    def parse_arr_to_sys(self, t : float, y : np.ndarray, vals : BGSystem):
        """
        Translate a GEF solution array to a BGSystem.

        Parameters
        ----------
        t : NDArray
            array of time coordinates
        y : NDArray
            array of variables at time t
        vals : BGSystem
            the target system
        """
        vals.units = False
        self.update_vals(t, y, vals)
        return
    
    def update_settings(self, **new_settings):
        """
        Update the `settings` of the class.
        """
        unknown_settings = []
        
        for setting, value in new_settings.items():
            if setting not in self.settings.keys():
                unknown_settings.append(setting)
            elif value != self.settings[setting]:
                print(f"Changing '{setting}' from {self.settings[setting]} to {value}.")
                self.settings[setting] = value
                if setting in ["atol", "rtol"]:
                    setattr(AuxTol, setting, value)
        
        if len(unknown_settings) > 0:
            print(f"Unknown settings: {unknown_settings}")
        return
    
    def _increase_ntr(self, val : int):
        self.ntr+=val
        print(f"Increasing 'ntr' by {val} to {self.ntr}.")
        return

    def _test_ode(self):
        try:
            t, yini, vals = self.initial_conditions()
        except Exception:
            raise ODECompilationError("Error while computing initial conditions.")
        try:
            self.ode(t, yini, vals)
        except Exception:
            raise ODECompilationError("Error while trying to compute a single ODE timestep.")
    
    
def GEFSolver(new_init : Callable, new_update_vals : Callable, new_timestep : Callable, new_variables : dict, new_events : list['Event'] = []):
    
    class CustomGEFSolver(BaseGEFSolver):
        vals_to_yini = staticmethod(new_init)
        update_vals = staticmethod(new_update_vals)
        timestep = staticmethod(new_timestep)
        known_variables = new_variables
        known_events = {event.name : event for event in new_events}
    
    CustomGEFSolver.__qualname__ = "GEFSolver"
    CustomGEFSolver.__module__ = __name__
    return CustomGEFSolver
        
    
class Event:
    def __init__(self, name : str, eventtype : str, func : Callable, terminal : bool, direction : int):
        """
        Initialise the event as `active`.

        Parameters
        ----------
        name : str
            sets the `name` attribute
        eventtype : str
            sets the `type` attribute
        func : Callable
            sets the `event_func` attribute
        terminal : boolean
            defines if the event occurrence is terminal or not
        direction : int
            defines the direction for which event occurrences are tracked
        """
        self.name : str= name
        """The name of the event."""

        self.type : str= eventtype
        """The event type 'terminal', 'error', or 'observer'."""
    
        self.active : bool = True
        """The events state, `False` implies the `Event` is disregarded by the solver."""

        func.terminal = terminal
        func.direction = direction
        
        self.event_func = func
        return
    
    @staticmethod
    def event_func(t : float, y : np.ndarray, sys : BGSystem) -> float:
        """
        The event tracked by `Event`

        This method is overwritten by the `func` input upon class initialisation.
        The signature and return of `func`needs to match this method

        Parameters
        ----------
        t : float
            time
        y : np.ndarray
            the solution array
        sys : BGSystem
            the system which is evolved alongside y

        Returns
        -------
        condition : float
            condition=0 is an event occurrence
        """
        return 1.
    
class TerminalEvent(Event):
    """
    An `Event` subclass whose occurrence terminates the solver.

    When the solver has terminated (due to an event or otherwise) it checks for `TerminalEvent` occurrences.
     This calls the `event_consequence` method, which returns instructions to `BaseGEFSolver`. 
     These instructions may be different depending on an event occurrence or a non-occurrence.
    """
    def __init__(self, name : str, func : Callable, direction : int, consequence : Callable):
        """
        Initialise the parent class and overwrite the `event_consequence` method.

        Parameters
        ----------
        name : str
            sets the `name` attribute
        func : Callable
            sets the `event_func` attribute
        direction : int
            passed as `direction` to the parent class
        consequence : Callable
            overwrites the `event_consequence` method
        """
        super().__init__(name, "terminal", func, True, direction)
        #ToDo: check for Consequence signature (once it is fixed)
        self.event_consequence = staticmethod(consequence)

    @staticmethod
    def event_consequence(sys : BGSystem, occurrence : bool) -> Tuple[str, dict]:
        """
        Inform the solver how to handle a (non-)occurrence of the event.
        
        This method is overwritten by the `consequence` input upon class initialisation.

        The methods returns are treated as an instruction to the solver:
        - `primary`: this informs the solver what to do with its ODE solution:
            - 'finish': the solver returns its solution marked as successful.
            - 'proceed': the solver continues solving from the termination time onwards.
            - 'repeat': the solver recomputes the solution.
        - `secondary`: this informs the solver if any of its settings need to be changed. 
        Allowed attributes are 'timestep', 'tend', 'atol', 'rtol'. See `BaseGEFSolver` for more details.

        Parameters
        ----------
        sys : BGSystem
            a system containing the solution of the solver
        occurrence : bool
            indicates if the event occurred during the solution or not

        Returns
        -------
        primary : str
            either 'finish', 'proceed' or 'repeat'
        secondary : dict
            the affected settings as keys and their new value as an item
        """
        if occurrence:
            return "proceed", {}
        else:
            return "proceed", {}


class ErrorEvent(Event):
    """
    An `Event` subclass whose occurrence indicates undesired behavior of the solution.

    When the solver terminates with an `ErrorEvent`, `BaseGEFSolver.solve_eom` returns the solution as unsuccessful.
    """
    def __init__(self, name : str, func : Callable, direction : int, message : str):
        """Initialise the parent class.
        
        The additional parameter `message` is printed on an event occurrence
        """
        super().__init__(name, "error", func, True, direction)
        self.message = message

class ObserverEvent(Event):
    """An `Event` which does not terminate the solve, but only records its occurences."""
    def __init__(self, name : str, func : Callable, direction : int):
        """Initialise the parent class."""
        super().__init__(name, "observer", func, False, direction)

class TruncationError(Exception):
    """
    Exception indicating that a GEF solution was unsuccessful.
    """
    pass

class ODECompilationError(Exception):
    """
    Exception indicating that an error occured while testing the ODE of a model.
    """
    pass

#define longer method docs from docs_solver:
generate_docs(docs_solver.DOCS)